/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	COMPILER DEBUGGING
 *
 *	Prints debugging information to STDOUT.
 *
 *	Public functions:
 *		DUMP()		Dump extended parse tree
 *		DUMPSYM()	Dump global symbol table
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "glue.h"

#include "proto.h"


/*
 *	DUMPNODE(NODE, DEPTH)
 *	Purpose:	Prints a node of the parse tree (menmonic, and
 *			optional parameter).
 *			Depth is indicated by indentation (using "| "
 *			and "|-").
 *	Receives:	NODE - pointer to node
 *			DEPTH - depth of NODE
 *	Returns:	-
 */
void dumpnode(node, depth)
struct node	*node;
int	depth;
{
	int	i;

	for (i=0; i<depth-1; i++) printf("| ");
	if (depth) printf("|-");

	if (node->token == GLUE) {
		printf("[]\n"); return;
	}

	if (node->token != G_NUM) printf("%s", tok2str(node->token));

	switch (node->token) {
		case G_LN: case G_LA: case G_LL: case G_LG: case G_LX:
		case G_SL: case G_SG: case G_SX: case G_LBL:
		case G_JMP: case G_JFA: case G_JTR: case G_SWI:
		case G_RMA: case G_VEC: case G_UNV: case G_NUM:
		case G_INI: case G_LIN:

			printf(" %d", node->v.val);
	}
	putchar('\n');
}


/*
 *	DUMPTREE(ROOT, DEPTH)
 *	Purpose:	Dumps a parse tree in prefix order.
 *	Receives:	ROOT - pointer to the root node
 *			DEPTH - depth of root node (usually 0)
 *	Returns:	-
 */
void dumptree(root, depth)
struct node	*root;
int	depth;
{
	if (!root) return;
	dumpnode(root, depth);
	dumptree(root->right, depth+1);
	dumptree(root->left, depth+1);
}


/*
 *	DUMP(ROOT)
 *	Purpose:	Dumps a parse tree
 *	Receives:	ROOT - pointer to the root node of the tree
 *	Returns:	-
 */
void dump(root)
struct node	*root;
{
	dumptree(root, 0);
}


/*
 *	DUMPSYM()
 *	Purpose:	Dumps the global symbol table.
 *			Output format is: FLAGS ADDR (HEX-ADDR) SIZE NAME
 *			Flags are:	Public
 *					Extern
 *					Static
 *					Forward	(should'nt happen)
 *					Rvalue
 *	Receives:	-
 *	Returns:	-
 */
void dumpsym() {
	int	i, j;
	struct syment	*s;

	putchar('\n');
	for (i=0; i<St0;) {
		if (Symbols[i].flags & STRUCT) {
			printf("STRUCT %s (%05d):\n", Symbols[i].name,
			Symbols[i].size);
			for (j=1; j<=Symbols[i].size; j++)
				printf("\t%05d %s\n", Symbols[i+j].addr,
				Symbols[i+j].name);
			i += j;
			continue;
		}
		s = &Symbols[i++];
		printf("%c%c%c%c%c %05d (0x%04x) %05d %s\n", 
		(s->flags&PUBLIC)? 'P': '-',
		(s->flags&EXTERN)? 'E': '-',
		(s->flags&STATIC)? 'S': '-',
		(s->flags&FORWARD)? 'F': '-',
		(s->flags&RVALUE)? 'R': '-',
		s->addr, s->addr, s->size, s->name);
	}
}
