///////////////////////////////////////////////////////////////////////////////
//
//	BSL/86 Minimum Runtime Support: Basic I/O
//	Copyright (C) 1995 Nils M. Holm
//

options "+O2";

#define(EXTERN(x),)	// Don't define the below functions as EXTERNs.
#include(blib.hdr)
#undefine(EXTERN)


//	READPACKED(FD, PSTR, LENGTH)
//	Writes the packed string PSTR to file FD.
//	LENGTH is the length of PSTR in CHARACTERS.
//
public readpacked(fd, pstr, length) {
	var	REGS.REGISTERS, end = length/_BPW_;

	REGS.R_AX = 0x3F00;
	REGS.R_BX = fd;
	REGS.R_CX = length;
	REGS.R_DX = pstr * _BPW_;
	pstr.end = pstr.(end+1) = 0;
	if (__SYS(_SYSINT_, REGS) & 1) return -1;
	return REGS.R_AX;
}


//	WRITEPACKED(FD, PSTR, LENGTH)
//	Writes the packed string PSTR to file FD.
//	LENGTH is the length of PSTR in CHARACTERS.
//
public writepacked(fd, pstr, length) {
	var	REGS.REGISTERS;

	REGS.R_AX = 0x4000;
	REGS.R_BX = fd;
	REGS.R_CX = length;
	REGS.R_DX = pstr * _BPW_;
	if (__SYS(_SYSINT_, REGS) & 1) return -1;
	return REGS.R_AX;
}


//	BLOCKREAD(FD, BLOCK, LENGTH)
//	Reads LENGTH words from file descriptor FD
//	to buffer at address BLOCK.
//	Returns the number of words read
//	(-1 means read error -- EOF?).
//
public blockread(fd, block, length) {
	var	rc;

	if ((rc = readpacked(fd, block, length*_BPW_)) == -1) return -1;
	return rc/_BPW_;
}


//	BLOCKWRITE(FD, BLOCK, LENGTH)
//	Writes LENGTH words starting at address BLOCK
//	to file descriptor FD.
//	Returns the number of words actually written
//	(-1 means write error -- disk full?).
//
public blockwrite(fd, block, length) {
	var	rc;

	if ((rc = writepacked(fd, block, length*_BPW_)) == -1) return -1;
	return rc/_BPW_;
}


//	PACK(S, D)
//	Packs the NUL-terminated array S into D.
//	On the i8086, this means that two
//	`characters' are packed into one word.
//	S and D may be the same.
//	PACK() returns the length of the packed string.
//
public pack(s, d) {
	var	i = 0, l = 0, c0, c1;

	while (1) {
		if (!(c0 = .s)) break;
		l += 1;
		c1 = s.1;
		d.i = c0 | c1<<8;
		i += 1;
		if (!c1) break;
		l += 1;
		s += 2;
	}
	d.i = 0;
	return l;
}


//	UNPACK(S, D)
//	Unpacks the packed string S into D.
//	On the i8086, this means that one
//	word is divided into two words with
//	the hibyte set to zero (characters).
//	S and D may be the same.
//	UNPACK() returns the length of the
//	unpacked array.
//
public unpack(s, d) {
	var	i = 0, c;

	while (1) {
		if (!(d.i = (c = .s) & 0xff)) break;
		i += 1;
		if (!(d.i = c>>8)) break;
		i += 1;
		s += 1;
	}
	return i+1;
}
