//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	COMPILER DEBUGGING
//
//	Prints debugging information to STDOUT.
//
//	Public functions:
//		DUMP()		Dump extended parse tree
//		DUMPSYM()	Dump global symbol table
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( glue.hdr)
#include( string.hdr)


extern	tok2str;


//	DUMPNODE(NP, DEPTH)
//	Purpose:	Prints a node of the parse tree (menmonic, and
//			optional parameter).
//			Depth is indicated by indentation (using "| "
//			and "|-").
//	Receives:	NP	- pointer to node
//			DEPTH	- depth of NP
//	Returns:	-
//
dumpnode(np, depth) {
	var	i;

	for (i=0, depth) writes(Outfile, ". ");

	if (np.N_token == GLUE) {
		writes(Outfile, "[]");
		writes(Outfile, _NLCHAR_);
		return;
	}

	if (np.N_token != G_NUM) writes(Outfile, tok2str(np.N_token));

	switch (np.N_token) {
 		case G_NUM:
 			writes(Outfile, "#");

		case G_LN: case G_LA: case G_LL: case G_LG: case G_LX:
		case G_SL: case G_SG: case G_SX: case G_LBL:
		case G_JMP: case G_JFA: case G_JTR: case G_SWI:
		case G_RMA: case G_VEC: case G_UNV:
		case G_INI: case G_LIN:

			writes(Outfile, "\s");
			writes(Outfile, itoa(np.N_val, 0));
	}
	writes(Outfile, _NLCHAR_);
}


//	DUMPTREE(ROOT, DEPTH)
//	Purpose:	Dumps a parse tree in prefix order.
//	Receives:	ROOT - pointer to the root node
//			DEPTH - depth of root node (usually 0)
//	Returns:	-
//
dumptree(root, depth) {
	if (!root) return;
	dumpnode(root, depth);
	dumptree(root.N_right, depth+1);
	dumptree(root.N_left, depth+1);
}


//	DUMP(ROOT)
//	Purpose:	Dumps a parse tree
//	Receives:	ROOT - pointer to the root node of the tree
//	Returns:	-
//
public dump(root) dumptree(root, 0);


//	DUMPSYM()
//	Purpose:	Dumps the global symbol table.
//			Output format is: FLAGS ADDR (HEX-ADDR) SIZE NAME
//			Flags are:	Public
//					Extern
//					Static
//					Forward	(should'nt happen)
//					Rvalue
//	Receives:	-
//	Returns:	-
//
public dumpsym() {
	var	i, j, s;
	var	buf.80;

	writes(Outfile, _NLCHAR_);
	ufor (i=0, St0, SYM) {
		if ((Symbols+i).S_flags & STCT) {
			strcpy(buf, "STRUCT ");
			strcat(buf, (Symbols+i).S_name);
			strcat(buf, " (");
			strcat(buf, itoa((Symbols+i).S_size, 5));
			strcat(buf, "):");
			strcat(buf, _NLCHAR_);
			writes(Outfile, buf);
			s = Symbols+i;
			ufor (j=SYM, s.S_size, SYM) {
				strcpy(buf, "\t");
				strcat(buf, itoa((s+j).S_addr, 5));
				strcat(buf, " ");
				strcat(buf, (s+j).S_name);
				strcat(buf, _NLCHAR_);
			}
			writes(Outfile, buf);
			i += j;
			loop;
		}

		s = Symbols+i;
		.buf =  (s.S_flags & PUB)-> 'P', '-';
		buf.1 = (s.S_flags & EXT)-> 'E', '-';
		buf.2 = (s.S_flags & STAT)-> 'S', '-';
		buf.3 = (s.S_flags & FORW)-> 'F', '-';
		buf.4 = (s.S_flags & RVAL)-> 'R', '-';
		buf.5 = '\s';
		buf.6 = 0;

		strcat(buf, itoa(s.S_addr, 5));
		strcat(buf, " ");
		strcat(buf, itoa(s.S_size, 5));
		strcat(buf, " ");
		strcat(buf, s.S_name);
		strcat(buf, _NLCHAR_);

		writes(Outfile, buf);
	}
}
