//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	THE SCANNER
//
//	The scanner reads the source program, and creates a token stream
//	for the parser.
//
//	Public functions:
//		SC_INIT()	Initialize the scanner
//		NEXT()		Read next token
//

#include( blib.hdr)

#include( token.hdr)
#include( scan.hdr)
#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( string.hdr)


var	Lastrd = 0;		// Result of last read operation


extern	report;


#define(isoct(c),	('0' <= (c) && (c) <= '7'))
#define(isdigit(c),	('0' <= (c) && (c) <= '9'))
#define(isalpha(c),	('a' <= (c) && (c) <= 'z' || 'A' <= (c) && (c) <= 'Z'))


//	SC_INIT()
//	Purpose:	Initialize the scanner.
//	Receives:	-
//	Returns:	-
//
public sc_init() {
	.Buffer = Bp = Btop = 0;
	Lastrd = 0;
	Infile = SYSIN;
	strcpy(Inname, "(SYSIN)");
}


//	READB()
//	Purpose:	Read a block from the source file.
//	Receives:	-
//	Returns:	0  - everything fine
//			-1 - EOF reached
//
readb() {
	var	xbuf.(BUFLEN+TOKLEN+2)/_BPW_;

	if (Lastrd < 0) {
		return (Bp >= Btop)-> -1, 0;
	}
	if (Bp != Btop && Btop > 0)
		bcopy(Buffer+Bp, Buffer, Btop-Bp);
	Bp = Btop-Bp;
	if ((Btop = readpacked(Infile, xbuf, BUFLEN)) < 0)
		report(EIO, FATAL);
	unpack(xbuf, Buffer+Bp);
	Buffer.(Btop+Bp) = 0;
	if (Btop < BUFLEN) Lastrd = -1;
	Btop += Bp;
	Bp = 0;
	return (!Btop)-> -1, 0;
}


//	SCOMMENT()
//	Purpose:	Skip a comment.
//	Receives:	-
//	Returns:	0  - end of comment found (OK)
//			-1 - EOF found in comment
//
scomment() {
	#gch(2);
	while (#ch() != '\n') {
		if (!#ch()) {
			if (Bp >= Btop-TOKLEN) {
				if (readb() < 0) return -1;
			}
		}
		#gch(1);
	}
	return 0;
}


//	SKIP()
//	Purpose:	Skip white space and comments.
//			Read next block, if necessary.
//	Receives:	-
//	Returns:	0  - everything fine
//			-1 - EOF found
//
skip() {
	var	c;

	while (1) {
		if (Bp >= Btop-TOKLEN) {
			if (readb() < 0) return -1;
		}
		if (#ch() == '/') {
			if (#nch() == '/') {
				if (scomment() < 0) return -1;
			}
		}
		c = #ch();
		if (!#iswhite(c)) break;
		if (c == '\n') Lno += 1;
		#gch(1);
	}
	return 0;
}


//	RCHE()
//	Purpose:	Read a (possibly escaped) character.
//			Convert escape sequences to ASCII values.
//	Receives:	-
//	Returns:	The ASCII value of the character.
//
rche() {
	var	v, i;

	if (#ch() != '\\') return #rch();

	#gch(1);
	switch(#ch()) {
		case 'A': case 'a':	#gch(1); return '\a';
		case 'B': case 'b':	#gch(1); return '\b';
		case 'E': case 'e':	#gch(1); return '\e';
		case 'F': case 'f':	#gch(1); return '\f';
		case 'N': case 'n':	#gch(1); return '\n';
		case 'R': case 'r':	#gch(1); return '\r';
		case 'S': case 's':	#gch(1); return '\s';
		case 'T': case 't':	#gch(1); return '\t';
		case 'V': case 'v':	#gch(1); return '\v';

		default:	if (#isoct(#ch())) {
					v = 0;
					for (i=0, 3) {
						if (!#isoct(#ch())) break;
						v = v*8 + (#ch()-'0');
						#gch(1);
					}
					return v;
				}
				return #rch();
	}
}


//	KEYWORD(S)
//	Purpose:	Check whether a string is a keyword.
//	Receives:	S	- the string
//	Returns:	The token associated with the keyword,
//			or a special failure value, if the string
//			does not contain a keyword.
//
keyword(s) {
	var	p;

	p = s;
	while (.p) {
		if ('a' <= .p && .p <= 'z') .p -= ('a'-'A');
		p += 1;
	}

	switch(.s) {
		case 'B':
			if (strequ(s, "BREAK")) return KBREAK;
			break;

		case 'C':
			if (strequ(s, "CASE")) return KCASE;
			break;

		case 'D':
			if (strequ(s, "DEFAULT")) return KDFLT;
			else if (strequ(s, "DO")) return KDO;
			break;

		case 'E':
			if (strequ(s, "ELSE")) return KELSE;
			else if (strequ(s, "EXTERN")) return KEXTERN;
			break;

		case 'F':
			if (strequ(s, "FOR")) return KFOR;
			else if (strequ(s, "FORWARD")) return KFORWD;
			break;

		case 'H':
			if (strequ(s, "HALT")) return KHALT;
			break;

		case 'I':
			if (strequ(s, "IF")) return KIF;
			break;

		case 'L':
			if (strequ(s, "LOOP")) return KLOOP;
			break;

		case 'O':
			if (strequ(s, "OPTIONS")) return KOPTS;
			break;

		case 'P':
			if (strequ(s, "PUBLIC")) return KPUBLIC;
			break;

		case 'R':
			if (strequ(s, "RETURN")) return KRETURN;
			break;

		case 'S':
			if (strequ(s, "STATIC")) return KSTATIC;
			else if (strequ(s, "STRUCTURE")) return KSTRUCT;
			else if (strequ(s, "SWITCH")) return KSWITCH;
			break;

		case 'U':
			if (strequ(s, "U")) return KUNSIGN;
			else if (strequ(s, "UFOR")) return KUFOR;
			else if (strequ(s, "UNSIGNED")) return KUNSIGN;
			else if (strequ(s, "UNTIL")) return KUNTIL;
			break;

		case 'V':
			if (strequ(s, "VAR")) return KVAR;
			else if (strequ(s, "VALUE")) return KVALUE;
			break;

		case 'W':
			if (strequ(s, "WHILE")) return KWHILE;
			break;
	}

	return FUCKUP;
}


//	SCANVAL()
//	Purpose:	Scan a numeric literal, auto-detect radix.
//			No overflow checking is done !!!
//	Receives:	-
//	Returns:	The value of the literal.
//
scanval() {
	var	radix, c, v, i=0;
	var	p, digits = "0123456789ABCDEF";
	var	d, o, x;

	if (#ch() == '0') {
		#gch(1);
		if (#ch() == 'x' || #ch() == 'X') {
			#gch(1); radix = 16;
		}
		else if (#ch() == 'b' || #ch() == 'B') {
			#gch(1); radix = 2;
		}
		else {
			radix = 8;
		}
	}
	else
		radix = 10;

	p = digits.radix;
	digits.radix = 0;

	x = 1; v = 0;
	while (1) {
		c = #ch();
		if ('a' <= c && c <= 'z') c -= ('a'-'A');
		if (!strchr(digits, c)) break;
		if ((i+=1) == TOKLEN) report(ETOKL, FATAL);
		o = v;
		d = c - (#isdigit(c)-> '0', ('A'-10));
		v = v*radix + d;
		if ((v-d)/radix != o && x) {
			report(EOVFL, WARNING); x = 0;
		}
		#gch(1);
	}
	#back();

	digits.radix = p;

	return v;
}


//	SCAN()
//	Purpose:	Extract a token from the source file.
//	Receives:	-
//	Returns:	The extracted token.
//
scan() {
	var	c;

	.Strval = Tvalue = 0;

	if (Btop < 0) return ENDFILE;
	if (#ch()) #gch(1);
	if (skip() < 0) {
		Btop = -1;
		return ENDFILE;
	}

	c = #ch();

	if (c == '.') return LV;
	else if (c == '@') return ADDROF;
	else if (c == ';') return SEMI;
	else if (c == ',') return COMMA;
	else if (c == '~') return NOT;
	else if (c == '(') return LPAREN;
	else if (c == ')') return RPAREN;
	else if (c == '[') return LBRACK;
	else if (c == ']') return RBRACK;
	else if (c == ':') return COLON;
	else if (c == '=') {
		if (#nch() == '=') {
			#gch(1); return EQU;
		}
		return MOV;
	}
	else if (c == '!') {
		if (#nch() == '=') {
			#gch(1); return NEQ;
		}
		return LNOT;
	}
	else if (c == '+') {
		if (#nch() == '=') {
			#gch(1); return AADD;
		}
		return ADD;
	}
	else if (c == '-') {
		if (#nch() == '=') {
			#gch(1); return ASUB;
		}
		else if (#nch() == '>') {
			#gch(1); return ARROW;
		}
		return SUB;
	}
	else if (c == '*') {
		if (#nch() == '=') {
			#gch(1); return AMUL;
		}
		return MUL;
	}
	else if (c == '/') {
		if (#nch() == '=') {
			#gch(1); return ADIV;
		}
		return DIV;
	}
	else if (c == '%') {
		if (#nch() == '=') {
			#gch(1); return AMOD;
		}
		return MOD;
	}
	else if (c == '&') {
		if (#nch() == '=') {
			#gch(1); return AAND;
		}
		else if (#nch() == '&') {
			#gch(1); return LAND;
		}
		return AND;
	}
	else if (c == '|') {
		if (#nch() == '=') {
			#gch(1); return AOR;
		}
		else if (#nch() == '|') {
			#gch(1); return LOR;
		}
		return OR;
	}
	else if (c == '^') {
		if (#nch() == '=') {
			#gch(1); return AXOR;
		}
		return XOR;
	}
	else if (c == '<') {
		if (#nch() == '=') {
			#gch(1); return LSEQ;
		}
		else if (#nch() == '<') {
			#gch(1);
			if (#nch() == '=') {
				#gch(1); return ASHL;
			}
			return SHL;
		}
		return LESS;
	}
	else if (c == '>') {
		if (#nch() == '=') {
			#gch(1); return GTEQ;
		}
		else if (#nch() == '>') {
			#gch(1);
			if (#nch() == '=') {
				#gch(1); return ASHR;
			}
			return SHR;
		}
		return GRTR;
	}
	else if (c == '#') {
		if (#nch() == '(') {
			#gch(1); return LBRACE;
		}
		else if (#nch() == ')') {
			#gch(1); return RBRACE;
		}
		if (#nch() == '<') {
			#gch(1); return LBRACK;
		}
		else if (#nch() == '>') {
			#gch(1); return RBRACK;
		}
		else {
			return HASH;
		}
	}
	else if (c == '{') {
		if (#nch() == '{') {
			#gch(1);
			if (#nch() == '=') {
				#gch(1); return AROL;
			}
			return ROL;
		}
		return LBRACE;
	}
	else if (c == '}') {
		if (#nch() == '}') {
			#gch(1);
			if (#nch() == '=') {
				#gch(1); return AROR;
			}
			return ROR;
		}
		return RBRACE;
	}

	else if (#isalpha(c) || c == '_') {
		var	i = 0;

		while (#isalpha(c) || c == '_' || #isdigit(c)) {
			if (i == TOKLEN) report(ETOKL, FATAL);
			if (i < TOKLEN-1) Strval.i = c; i+=1;
			#gch(1);
			c = #ch();
		}
		Strval.i = 0;
		#back();

		if ((i = keyword(Strval)) != FUCKUP) return i;

		Strval.NAMELEN = 0;
		return SYMBOL;
	}

	else if (c == '\'') {
		#gch(1);
		Tvalue = rche();
		if (#ch() != '\'') report(ETERM, ERROR);
		return NUMBER;
	}

	else if (#isdigit(c)) {
		Tvalue = scanval();
		return NUMBER;
	}

	// The stupid preprocessor can only handle balanced quotes,
	// so DO NOT remove the below comments!
	else if (c == '"') {		// " DO NOT DELETE
		var	i = 0;

		#gch(1);
		c = #ch();
		while (c != '"') {	// " DO NOT DELETE
			if (c == '\n') report(ETERM, ERROR);
			if (i >= TOKLEN) report(ETOKL, FATAL);
			Strval.i = rche(); i+=1;
			c = #ch();
		}
		Strval.i = 0;
		return STRING;
	}

	else {
		report(ESCAN, FATAL);
	}

	return FUCKUP;
}


//	NEXT()
//	Purpose:	Read the next token from the input stream.
//	Receives:	-
//	Returns:	The token read.
//
public next() {
	return (Token = scan());
}
