///////////////////////////////////////////////////////////////////////////////
//
//	BSL COMPILER VALIDATION SUIT
//	COPYRIGHT (C) 1995 NILS M. HOLM
//
//	TO RUN THE TEST, COMPILE AND LAUNCH THIS PROGRAM.
//
//	IT SHOULD COMPILE WITH ONE WARNING MESSAGE (LOCAL
//	VARIABLE SHADOWS GLOBAL).
//	WHEN RUN, THE PROGRAM REPORTS EACH FAILED TEST TO
//	SYSOUT. WHEN THE TEST PASSES, IT SIMPLY PRINTS
//	THE NUMBER OF TEST PERFORMED.
//
//	THE TEST SUIT REQUIRES (BESIDES A BSL COMPILER)
//	THE MACRO PROCESSOR `MPP', AND THE RUNTIME SUPPORT
//	FUNCTIONS `PACK()' AND `WRITEPACKED()'. `MPP' IS JUST
//	USED TO INCLUDE THE EXTERN DECLS FOR THE ABOVE FUNTIONS.
//
//	_BPW_ (BYTES PER WORD) MUST BE DEFINED PROPERLY TO
//	PASS THIS TEST.
//
//	SOME TESTS WILL ONLY PASS ON TWO'S COMPLEMENT MACHINES,
//	BUT I DON'T CONSIDER THIS A REAL BIG PROBLEM.
//
//	THE BSL COMPILER VALIDATION SUIT ASSUMES THAT THE
//	COMPILER *WORKS*. IT IS A POOR TOOL FOR *DETECTING*,
//	OR EVEN REPORTING ERRORS.
//
//	SOME OF THE BELOW TESTS USUALLY CANNOT FAIL (LIKE
//	ASSOCIATIVITY TESTS WITH `+'), BUT YOU NEVER KNOW ...
//

#include( blib.hdr)

options "+O2";	// ALSO RUN THE TEST WITH OPTIMIZATIONS DISABLED !!
options "+L";


value	BUFLEN = 256/_BPW_;

var	TestID = 0,
	TestN = 0,
	ErrorC = 0;


writes(s) {
	var	buf.BUFLEN, i;

	while (s.i) i+=1;
	pack(s, buf);
	writepacked(1, buf, i);
}


writen(n) {
	var	buf.16, i = 0, j, k, t;

	while (n) {
		buf.i = n%10+'0';
		i += 1;
		n /= 10;
	}
	buf.i = 0;
	k =  i/2;
	i -= 1;
	for (j=0, k) {
		t = buf.j;
		buf.j = buf.(i-j);
		buf.(i-j) = t;
	}
	writes(buf);
}


test(a, b) {
	if (a != b) {
		writes("FAILED: #");
		writen(TestID);
		writes("\r\n");
		ErrorC += 1;
	}
	TestID += 1;
	TestN += 1;
}


fail() test(0, !0);


///////////////////////////////////////////////////////////////////////////////
//
//	GENERAL OPERATOR TESTS
//
oper_tests() {
	var	a = 100;

	TestID = 100;

	test((10, 25), 25);

	test(a += 100, 200);
	test(a -= 100, 100);
	test(a *= 5,   500);
	test(a /= 2,   250);
	test(a %= 99,   52);
	test(a <<= 3,  416);
	test(a >>= 2,  104);
	test(a {{= 1,  208);
	test(a }}= 1,  104);
	a = 0x0F0F;
	test(a |= 0xF0,  0xFFF);
	test(a &= 0xF00, 0xF00);
	test(a ^= 0xFFF, 0x0FF);

	a = 0xF0F;
	test(a | 0xAAA, 0xFAF);
	test(a ^ 0xEEE, 0x1E1);
	test(a & 0x333, 0x303);

	test(a != 500,  -1);
	test(a == 500,   0);
	a = -1;
	test(a <  1000, -1);
	test(a >  1000,  0);
	test(a <= 1000, -1);
	test(a >= 1000,  0);
	test(a UNSIGNED <  1000,  0);
	test(a UNSIGNED >  1000, -1);
	test(a UNSIGNED <= 1000,  0);
	test(a UNSIGNED >= 1000, -1);

	a = 100;
	test(a + 100,   200);
	test(a - 100,     0);
	test(a * 100, 10000);
	test(a /  10,    10);
	test(a %   9,     1);

	test(+a, +100);
	test(-a, -100);
	test(~a & 0xff, 0x9b);
	test(!a,    0);
	test(!!a,  -1);

	test(5*(5+5)*5, 250);
}


///////////////////////////////////////////////////////////////////////////////
//
//	OPERATOR PRECEDENCE TESTS
//
oper_p1() {
	var	a, b, c, d;

	TestID = 200;

	test((a = 100, b = 200), ((a = 100), (b = 200)));
	c = a; b = d;
	test((a &= 555, b &= 555), ((c &= 555), (d &= 555)));
	test((a |= 255, b |= 255), ((c |= 255), (d |= 255)));
	test((a ^= 205, b ^= 205), ((c ^= 205), (d ^= 205)));
	test((a <<= 3, b <<= 3), ((c <<= 3), (d <<= 3)));
	test((a >>= 3, b >>= 3), ((c >>= 3), (d >>= 3)));
	test((a {{= 3, b {{= 3), ((c {{= 3), (d {{= 3)));
	test((a }}= 3, b }}= 3), ((c }}= 3), (d }}= 3)));
	test((a += 100, b += 200), ((c += 100), (d += 200)));
	test((a += 100, b += 200), ((c += 100), (d += 200)));
	test((a -= 100, b -= 200), ((c -= 100), (d -= 200)));
	test((a *= 10,  b *= 20),  ((c *= 10),  (d *= 20)));
	test((a /= 10,  b /= 20),  ((c /= 10),  (d /= 20)));
	test((a %= 77,  b %= 88),  ((c %= 77),  (d %= 88)));
}


oper_p2() {
	var	a, b;

	TestID = 300;

	test(a = 100 | 200, 100 | 200);
	b = a;
	test(a &= 222 | 333, b &= (222 | 333));
	test(a |= 127 | 128, b |= (127 | 128));
	test(a ^= 200 | 005, b ^= (200 | 005));
	test(a <<= 1 | 2,    b <<= (1 | 2));
	test(a >>= 1 | 2,    b >>= (1 | 2));
	test(a {{= 1 | 2,    b {{= (1 | 2));
	test(a }}= 1 | 2,    b }}= (1 | 2));
	test(a += 100 | 200, b += (100 | 200));
	test(a -= 100 | 200, b -= (100 | 200));
	test(a *= 10  | 20,  b *= (10  | 20));
	test(a /= 10  | 20,  b /= (10  | 20));
	test(a %= 77  | 55,  b %= (77  | 55));
}


oper_p3() {
	var	a;

	TestID = 400;

	a = 1234;
	test(a | 999 ^ 777,  a | (999 ^ 777));
	test(a ^ 999 & 777,  a ^ (999 & 777));
	test(a & 123 != 456, a & (123 != 456));
	test(a & 123 == 456, a & (123 == 456));

	test(a == a <  9999, a == (a <  9999));
	test(a == a >  1111, a == (a >  1111));
	test(a == a <= 1111, a == (a <= 1111));
	test(a == a >= 9999, a == (a >= 9999));
	test(a == a U<  1111, a == (a U<  1111));
	test(a == a U>  9999, a == (a U>  9999));
	test(a == a U<= 9999, a == (a U<= 9999));
	test(a == a U>= 1111, a == (a U>= 1111));
}


oper_p4() {
	var	a = 4032;

	TestID = 500;

	test(a <  a >> 3, a <  (a >> 3));
	test(a >  a >> 3, a >  (a >> 3));
	test(a <= a >> 3, a <= (a >> 3));
	test(a >= a >> 3, a >= (a >> 3));
	test(a U<  a >> 3, a U<  (a >> 3));
	test(a U>  a >> 3, a U>  (a >> 3));
	test(a U<= a >> 3, a U<= (a >> 3));
	test(a U>= a >> 3, a U>= (a >> 3));

	test(a <  a << 3, a <  (a << 3));
	test(a >  a << 3, a >  (a << 3));
	test(a <= a << 3, a <= (a << 3));
	test(a >= a << 3, a >= (a << 3));
	test(a U<  a << 3, a U<  (a << 3));
	test(a U>  a << 3, a U>  (a << 3));
	test(a U<= a << 3, a U<= (a << 3));
	test(a U>= a << 3, a U>= (a << 3));

	test(a <  a }} 3, a <  (a }} 3));
	test(a >  a }} 3, a >  (a }} 3));
	test(a <= a }} 3, a <= (a }} 3));
	test(a >= a }} 3, a >= (a }} 3));
	test(a U<  a }} 3, a U<  (a }} 3));
	test(a U>  a }} 3, a U>  (a }} 3));
	test(a U<= a }} 3, a U<= (a }} 3));
	test(a U>= a }} 3, a U>= (a }} 3));

	test(a <  a {{ 3, a <  (a {{ 3));
	test(a >  a {{ 3, a >  (a {{ 3));
	test(a <= a {{ 3, a <= (a {{ 3));
	test(a >= a {{ 3, a >= (a {{ 3));
	test(a U<  a {{ 3, a U<  (a {{ 3));
	test(a U>  a {{ 3, a U>  (a {{ 3));
	test(a U<= a {{ 3, a U<= (a {{ 3));
	test(a U>= a {{ 3, a U>= (a {{ 3));
}


oper_p5() {
	var	a;

	TestID = 600;

	a = -12345;
	test(a << 2 + 1, a << (2 + 1));
	test(a >> 2 + 1, a >> (2 + 1));
	test(a {{ 2 + 1, a {{ (2 + 1));
	test(a }} 2 + 1, a }} (2 + 1));
	test(a << 7 - 3, a << (7 - 3));
	test(a >> 7 - 3, a >> (7 - 3));
	test(a {{ 7 - 3, a {{ (7 - 3));
	test(a }} 7 - 3, a }} (7 - 3));

	a = 4321;
	test(a + 10 * 20,  a + (10 * 20));
	test(a - 10 * 20,  a - (10 * 20));
	test(a + 200 / 7,  a + (200 / 7));
	test(a - 200 / 7,  a - (200 / 7));
	test(a + 999 % 11, a + (999 % 11));
	test(a - 999 % 11, a - (999 % 11));
}


oper_p6() {
	var	a = 5432;

	TestID = 700;

	test(+a   * 10, (+a)   * 10);
	test(++a  * 20, (++a)  * 20);
	test(+++a * 30, (+++a) * 30);
	test(-a   * 11, (-a)   * 11);
	test(--a  * 22, (--a)  * 22);
	test(---a * 33, (---a) * 33);
	test(~a   * 10, (~a)   * 10);
	test(~~a  * 20, (~~a)  * 20);
	test(~~~a * 30, (~~~a) * 30);
	test(!a   * 11, (!a)   * 11);
	test(!!a  * 22, (!!a)  * 22);
	test(!!!a * 33, (!!!a) * 33);

	test(+a   / 10, (+a)   / 10);
	test(++a  / 20, (++a)  / 20);
	test(+++a / 30, (+++a) / 30);
	test(-a   / 11, (-a)   / 11);
	test(--a  / 22, (--a)  / 22);
	test(---a / 33, (---a) / 33);
	test(~a   / 10, (~a)   / 10);
	test(~~a  / 20, (~~a)  / 20);
	test(~~~a / 30, (~~~a) / 30);
	test(!a   / 11, (!a)   / 11);
	test(!!a  / 22, (!!a)  / 22);
	test(!!!a / 33, (!!!a) / 33);

	test(+a   % 10, (+a)   % 10);
	test(++a  % 20, (++a)  % 20);
	test(+++a % 30, (+++a) % 30);
	test(-a   % 11, (-a)   % 11);
	test(--a  % 22, (--a)  % 22);
	test(---a % 33, (---a) % 33);
	test(~a   % 10, (~a)   % 10);
	test(~~a  % 20, (~~a)  % 20);
	test(~~~a % 30, (~~~a) % 30);
	test(!a   % 11, (!a)   % 11);
	test(!!a  % 22, (!!a)  % 22);
	test(!!!a % 33, (!!!a) % 33);
}


prec_tests() {
	oper_p1();
	oper_p2();
	oper_p3();
	oper_p4();
	oper_p5();
	oper_p6();
}


///////////////////////////////////////////////////////////////////////////////
//
//	OPERATOR ASSOCIATIVITY TESTS
//
oper_a1() {
	var	a, b, c, d;

	TestID = 800;

	a = b = c = d = -1;
	test(a = b = 1234, c = (d = 1234));
	test(a += b += 1234, c += (d += 1234));
	test(a -= b -= 1234, c -= (d -= 1234));
	test(a *= b *= 3, c *= (d *= 3));
	a = c = 43210;
	test(a /= b /= 3, c /= (d /= 3));
	a = c = 43210;
	test(a %= b %= 999, c %= (d %= 999));
	a = c = 12345;
	test(a &= b &= 9999, c &= (d &= 9999));
	test(a |= b |= 6144, c |= (d |= 6144));
	test(a ^= b ^= -1,   c ^= (c ^= -1));
	a = c = 1234; b = d = 1;
	test(a <<= b <<= 1, c <<= (d <<= 1));
	test(a >>= b >>= 1, c >>= (d >>= 1));
	test(a }}= b }}= 1, c }}= (d }}= 1));
	test(a {{= b {{= 1, c {{= (d {{= 1));
}


oper_a2() {
	var	a = -1234;

	TestID = 900;

	test((a, 789, 345), 345);

	test(a | 0xff | 500, (a | 0xff) | 500);
	test(a & 0xff & 500, (a & 0xff) & 500);
	test(a ^ 0xff ^ 500, (a ^ 0xff) ^ 500);

	test(a == -1234 == -1, (a == -1234) == -1);
	test(a != -1234 != 1, (a != -1234) != 1);

	test(a >  500 >  0,   (a >  500) >  0);
	test(a <  500 <  0,   (a <  500) <  0);
	test(a >= 500 >= 0,   (a >= 500) >= 0);
	test(a <= 500 <= 0,   (a <= 500) <= 0);
	test(a U>  500 U>  0, (a U>  500) U>  0);
	test(a U<  500 U<  0, (a U<  500) U<  0);
	test(a U>= 500 U>= 0, (a U>= 500) U>= 0);
	test(a U<= 500 U<= 0, (a U<= 500) U<= 0);

	test(a >> 2 >> 3, (a >> 2) >> 3);
	test(a << 2 << 3, (a << 2) << 3);
	test(a }} 2 }} 3, (a }} 2) }} 3);
	test(a {{ 2 {{ 3, (a {{ 2) {{ 3);

	test(a + 123 + 456, (a + 123) + 456);
	test(a - 123 - 456, (a - 123) - 456);

	test(a * 12 * 5, (a * 12) * 5);
	test(a / 12 / 5, (a / 12) / 5);
	test(a % 99 % 7, (a % 99) % 7);

	test(++a, +(+a));
	test(--a, -(-a));
	test(~~a, ~(~a));
	test(!!a, !(!a));
}


assoc_tests() {
	oper_a1();
	oper_a2();
}


///////////////////////////////////////////////////////////////////////////////
//
//	CHECK EQUIVALENCE OF CONSTANT EXPRESSIONS AND GENERAL EXPRESSIONS
//

value	C001 = 5000 | 0xf0f;
value	C002 = 5000 & 0xf0f;
value	C003 = 5000 ^ 0xf0f;
value	C004 = 5000 << 2;
value	C005 = 5000 >> 2;
value	C006 = 5000 {{ 3;
value	C007 = 5000 + 12345;
value	C008 = 5000 - 12345;
value	C009 = 5000 * 9;
value	C010 = 5000 / 12;
value	C011 = 5000 % 999;
value	C012 = +5000;
value	C013 = -5000;
value	C014 = ~5000;
value	C015 = !5000;
value	C016 = !0;

cexpr_oper() {
	var	a = 5000;

	TestID = 1000;

	test(C001, a | 0xf0f);
	test(C002, a & 0xf0f);
	test(C003, a ^ 0xf0f);
	test(C004, a << 2);
	test(C005, a >> 2);
	test(C006, a {{ 3);
	test(C007, a + 12345);
	test(C008, a - 12345);
	test(C009, a * 9);
	test(C010, a / 12);
	test(C011, a % 999);
	test(C012, +a);
	test(C013, -a);
	test(C014, ~a);
	test(C015, !a);
	a = 0;
	test(C016, !a);
}


value	P001 = 5000 | 1234 ^ 4321,
	P002 = 5000 ^ 1234 & 1230,
	P003 = 5000 & 1234 << 1,
	P004 = 5000 & 1234 >> 1,
	P005 = 5000 & 1234 {{ 1,
	P006 = 5000 & 1234 }} 1,
	P007 = 5000 << 1 + 2,
	P008 = 5000 << 3 - 1,
	P009 = 5000 >> 1 + 2,
	P010 = 5000 >> 3 - 1,
	P011 = 5000 {{ 1 + 2,
	P012 = 5000 {{ 3 - 1,
	P013 = 5000 }} 1 + 2,
	P014 = 5000 }} 3 - 1,
	P015 = 5000 + 12 * 34,
	P016 = 5000 + 3456 / 12,
	P017 = 5000 - 12 * 34,
	P018 = 5000 - 3456 / 12;

cexpr_prec() {
	var	a = 5000;

	TestID = 1100;

	test(P001, a | 1234 ^ 4321);
	test(P002, a ^ 1234 & 1230);
	test(P003, a & 1234 << 1);
	test(P004, a & 1234 >> 1);
	test(P005, a & 1234 {{ 1);
	test(P006, a & 1234 }} 1);
	test(P007, a << 1 + 2);
	test(P008, a << 3 - 1);
	test(P009, a >> 1 + 2);
	test(P010, a >> 3 - 1);
	test(P011, a {{ 1 + 2);
	test(P012, a {{ 3 - 1);
	test(P013, a }} 1 + 2);
	test(P014, a }} 3 - 1);
	test(P015, a + 12 * 34);
	test(P016, a + 3456 / 12);
	test(P017, a - 12 * 34);
	test(P018, a - 3456 / 12);
}


value	A001 = 5000 |  123 |  456,
	A002 = 5000 ^ 4095 ^ 2000,
	A003 = 5000 & 0xf00 & 0xfff,
	A004 = 5000 >> 1 >> 2,
	A005 = 5000 << 1 << 2,
	A006 = 5000 }} 1 }} 2,
	A007 = 5000 {{ 1 {{ 2,
	A008 = 5000 + 456 + 789,
	A009 = 5000 - 456 - 789,
	A010 = 5000 * 2 * 3,
	A011 = 5000 / 3 / 2,
	A012 = 5000 % 999 % 77,
	A013 = +5000,
	A014 = ++5000,
	A015 = +++5000,
	A016 = -5000,
	A017 = --5000,
	A018 = ---5000,
	A019 = ~5000,
	A020 = ~~5000,
	A021 = ~~~5000,
	A022 = !5000,
	A023 = !!5000,
	A024 = !!!5000;

cexpr_assoc() {
	var	a = 5000;

	TestID = 1200;

	test(A001, a |  123 |  456);
	test(A002, a ^ 4095 ^ 2000);
	test(A003, a & 0xf00 & 0xfff);
	test(A004, a >> 1 >> 2);
	test(A005, a << 1 << 2);
	test(A006, a }} 1 }} 2);
	test(A007, a {{ 1 {{ 2);
	test(A008, a + 456 + 789);
	test(A009, a - 456 - 789);
	test(A010, a * 2 * 3);
	test(A011, a / 3 / 2);
	test(A012, a % 999 % 77);
	test(A013, +a);
	test(A014, ++a);
	test(A015, +++a);
	test(A016, -a);
	test(A017, --a);
	test(A018, ---a);
	test(A019, ~a);
	test(A020, ~~a);
	test(A021, ~~~a);
	test(A022, !a);
	test(A023, !!a);
	test(A024, !!!a);
}


cexpr_tests() {
	cexpr_oper();
	cexpr_prec();
	cexpr_assoc();
}


///////////////////////////////////////////////////////////////////////////////
//
//	SIGN TESTS
//

sign_tests() {
	var	a;

	TestID = 1300;

	a = -1;
	test(a < 0,  -1);
	test(a <= 0, -1);
	test(a > 0,  0);
	test(a >= 0, 0);
	test(a UNSIGNED < 0,  0);
	test(a UNSIGNED <= 0, 0);
	test(a UNSIGNED > 0,  -1);
	test(a UNSIGNED >= 0, -1);
	test(a UNSIGNED < 0,  a U<  0);
	test(a UNSIGNED <= 0, a U<= 0);
	test(a UNSIGNED > 0,  a U>  0);
	test(a UNSIGNED >= 0, a U>= 0);
}


///////////////////////////////////////////////////////////////////////////////
//
//	ADDRESSING TESTS
//

var	G000;

atom_tests() {
	static var	S001;
	var		L001;
	var		globp, locp, statp;

	TestID = 1400;

	globp = @G000;
	statp = @S001;
	locp = @L001;
	G000 = 12345; test(G000, .globp);
	S001 = 4321;  test(S001, .statp);
	L001 = 21212; test(L001, .locp);
	.globp = -9999; test(.globp, G000);
	.statp = -7777; test(.statp, S001);
	.locp = -5555; test(.locp, L001);
}


value	SIZE = 100;
var	GV00.SIZE;

vector_tests() {
	static var	SV01.SIZE;
	var	LV01.SIZE, i, j;

	TestID = 1500;

	j = 1;
	for (i=0, SIZE) GV00.i = (j *= 123);
	for (i=0, SIZE) SV01.i = GV00.i;

	TestN += 1;
	TestID += 1;
	j = 1;
	for (i=0, SIZE) if (SV01.i != (j *= 123)) {
		fail(); break;
	}

	for (i=0, SIZE) LV01.i = SV01.i;

	TestN += 1;
	TestID += 1;
	j = 1;
	for (i=0, SIZE) if (LV01.i != (j *= 123)) {
		fail(); break;
	}

	j = 1;
	for (i=0, SIZE) .(GV00+i) = (j *= 123);
	for (i=0, SIZE) .(SV01+i) = .(GV00+i);

	TestN += 1;
	TestID += 1;
	j = 1;
	for (i=0, SIZE) if (.(SV01+i) != (j *= 123)) {
		fail(); break;
	}

	for (i=0, SIZE) .(LV01+i) = .(SV01+i);

	TestN += 1;
	TestID += 1;
	j = 1;
	for (i=0, SIZE) if (.(LV01+i) != (j *= 123)) {
		fail(); break;
	}

	TestN += 1;
	TestID += 1;
	for (i=0, SIZE) GV00.i = i;
	LV01 = GV00;
	for (i=0, SIZE) if (LV01.i != i) {
		fail(); break;
	}
}


struct_tests() {
	structure T1 {
		S_x, S_y, S_z
	};
	structure T2 {
		S_a, S_b, S_c.T1, S_d.100
	};
	var	s1.T1, s2.T2;
	var	i;

	TestID = 1600;

	test(T1, 3);
	test(T2, 2+T1+100);
	test(S_x, 0);
	test(S_y, 1);
	test(S_z, 2);
	test(S_a, 0);
	test(S_b, 1);
	test(S_c, 2);
	test(S_d, 2+T1);

	s1.S_x = 123; test(s1.S_x, .s1);
	s1.S_y = 456; test(s1.S_y, s1.1);
	s1.S_z = 789; test(s1.S_z, s1.2);

	i = s2 + S_c;
	i.S_x = 12345;
	test(i.S_x, s2.(S_c+S_x));
}


var	glob;

local_tests() {
	var	loc = 100;

	TestID = 1700;

	glob = 100;
	{
		var	loc = 200;
		{
			var	loc = 300;
			{
				var	loc = 400;
				{
					var	loc = 500;
					glob += loc;
				}
				glob += loc;
			}
			glob += loc;
		}
		glob += loc;
	}
	glob += loc;
	test(glob, 1600);

	{
		var	glob;
		glob = -1;
	}
	test(glob, 1600);
}


array_tests() {
	var	V;
	value	CONSTVAL = 12;

	TestID = 1800;

	V = [ 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096 ];
	test(V.0,  .V);
	test(V.0,  2);
	test(V.1,  4);
	test(V.2,  8);
	test(V.3,  16);
	test(V.4,  32);
	test(V.5,  64);
	test(V.6,  128);
	test(V.7,  256);
	test(V.8,  512);
	test(V.9,  1024);
	test(V.10, 2048);
	test(V.11, 4096);

	V = "HELLO, WORLD!";
	test(V.0,  'H');
	test(V.1,  'E');
	test(V.2,  'L');
	test(V.3,  'L');
	test(V.4,  'O');
	test(V.5,  ',');
	test(V.6,  '\S');
	test(V.7,  'W');
	test(V.8,  'O');
	test(V.9,  'R');
	test(V.10, 'L');
	test(V.11, 'D');
	test(V.12, '!');
	test(V.13, 0);

	V = [ CONSTVAL * 12, @array_tests, "TEXT" ];
	test(.V, CONSTVAL * 12);	//1827
	test(V.1, @array_tests);
	V = .(V+2);
	test(V.0, 'T');
	test(V.1, 'E');
	test(V.2, 'X');
	test(V.3, 'T');
	test(V.4, 0);
}


nesting_tests() {
	var	ntable;

	TestID = 1900;

	ntable = [
		[ 00, 01, 02 ],
		[ 10, 11, 12 ],
		[ 20, 21, 22 ]
	];

	test(ntable.0.0, 00);
	test(ntable.0.1, 01);
	test(ntable.0.2, 02);

	test(ntable.1.0, 10);
	test(ntable.1.1, 11);
	test(ntable.1.2, 12);

	test(ntable.2.0, 20);
	test(ntable.2.1, 21);
	test(ntable.2.2, 22);
}


addr_tests() {
	atom_tests();
	vector_tests();
	struct_tests();
	local_tests();
	array_tests();
	nesting_tests();
}


///////////////////////////////////////////////////////////////////////////////
//
//	FUNCTION TESTS
//

fn(n) return n;


fn_addv(v) {
	var	vp = @v, sum = 0;

	while (.vp) {
		sum += .vp;
		vp += 1;
	}
	return sum;
}


forward	fac2;

fac(n) return n<2-> 1, n*fac2(n);

fac2(n) return fac(n-1);


func_tests() {
	var	fp;

	TestID = 2000;

	test(fn(12345), 12345);
	test(fn(123 + 456), 123 + 456);
	test(fn(12 * 34 + 56 << 2), 12 * 34 + 56 << 2);
	test(fn(fn(fn(123) + fn(456))), 123 + 456);

	test(fn_addv(10, 20, 0), 30);
	test(fn_addv(1, 2, 3, 4, 5, 6, 7, 8, 9, 0), 45);
	test(fn_addv(0), 0);

	test(fac(0), 1);
	test(fac(1), 1);
	test(fac(2), 2);
	test(fac(3), 6);
	test(fac(4), 24);
	test(fac(5), 120);
	test(fac(6), 720);
	test(fac(7), 5040);

	fp = fn;
	test(fp(129), 129);

	fp = fn_addv;
	test(fp(123, 456, 789, 0), 1368);

	fp = fac;
	test(fp(7), 5040);
}


///////////////////////////////////////////////////////////////////////////////
//
//	BRANCH TESTS
//

if_test(a, b, c, d, e) {
	var	r;

	if (a<b)
		if (c<d)
			r = 1;
		else
			r = 2;
	else
		if (c<d)
			r = 3;
		else
			r = 4;

	test(r, e);
}


branch_tests() {
	TestID = 2100;

	if_test(0, 1, 0, 1, 1);
	if_test(0, 1, 1, 0, 2);
	if_test(1, 0, 0, 1, 3);
	if_test(1, 0, 1, 0, 4);
}


test_switch(n) switch(n) {
	case 1:	return 111;
	case 2:	return 222;
	case 3:	return 333;
	case 4:	return 444;
	case 5:	return 555;
	default:return(-1);
}


switch_tests() {
	var	a;

	TestID = 2200;

	a = 1000;
	switch (a) {
		case 4000:	a = 5000; loop;
		case 1000:	a = 2000; loop;
		case 2000:	a = 3000;
		case 3000:	a = 4000; loop;
		case -1:	a = 999; break;
		case 5000:	a = -1; loop;
	}

	test(a, 999);

	test(test_switch(1), 111);
	test(test_switch(2), 222);
	test(test_switch(3), 333);
	test(test_switch(4), 444);
	test(test_switch(5), 555);
	test(test_switch(9999), -1);
	test(test_switch(0), -1);
	test(test_switch(-1), -1);
}


///////////////////////////////////////////////////////////////////////////////
//
//	LOOP TESTS
//

while_test() {
	var	a, b;

	TestID = 2300;

	a = 0;
	while (a<100) a+=1;
	test(a, 100);

	a = b = 0;
	while (a<100) {
		b += 12;
		a += 1;
	}
	test(b, 1200);

	b = a = 0;
	while (1) {
		if ((a += 1) >= 100) break;
		if (!(a%2)) loop;
		b += 1;
	}
	test(b, 50);

	a = 0;
	while (1) {
		a = 1; break;
	}
	test(a, 1);

	a = 0;
	while (0) {
		a = 1;
	}
	test(a, 0);
}


until_test() {
	var	a, b;

	TestID = 2400;

	a = 0;
	do a += 1; until(a >= 100);
	test(a, 100);

	a = b = 0;
	do {
		b += 25;
		a += 1;
	} until(a >= 100);
	test(b, 2500);

	b = a = 0;
	do {
		if ((a += 1) >= 100) break;
		if (!(a%2)) loop;
		b += 1;
	} until(0);
	test(b, 50);

	a = 0;
	do {
		a = 1; break;
	} until(0);
	test(a, 1);

	a = 0;
	do {
		a = 1;
	} until(1);
	test(a, 1);
}


for_test() {
	var	a, b;

	TestID = 2500;

	for (a=0, 100)
		;
	test(a, 100);

	for(a = b=0, 100) b += 123;
	test(a, 100);
	test(b, 12300);

	for (a=b=0, 999, 0) {
		if ((a += 1) >= 100) break;
		if (!(a%2)) loop;
		b += 1;
	}
	test(b, 50);

	for (a=0, 100, 0) a += 1;
	test(a, 100);

	b = 0;
	for (a=0, 0, 0) b=1;
	test(b, 0);

	b = 0;
	for (a=0, 1, 0) {
		b = 1; break;
	}
	test(b, 1);

	for (a=-10, 0)
		;
	test(a, 0);
}


ufor_test() {
	var	a, b;

	TestID = 2600;

	ufor (a=0, 100)
		;
	test(a, 100);

	ufor(a = b=0, 100) b += 123;
	test(a, 100);
	test(b, 12300);

	ufor (a=b=0, 999, 0) {
		if ((a += 1) >= 100) break;
		if (!(a%2)) loop;
		b += 1;
	}
	test(b, 50);

	ufor (a=0, 100, 0) a += 1;
	test(a, 100);

	b = 0;
	ufor (a=0, 0, 0) b=1;
	test(b, 0);

	b = 0;
	ufor (a=0, 1, 0) {
		b = 1; break;
	}
	test(b, 1);

	b = 0;
	ufor (a=-11, -1) b += 1;
	test(a, -1);
	test(b, 10);
}


loop_tests() {
	while_test();
	until_test();
	for_test();
	ufor_test();
}


///////////////////////////////////////////////////////////////////////////////
//
//	INITIALIZATION TESTS
//

var	GI00 = 5*(5+5)*5;

init() return 5432;

s_init() {
	static var	S = 777;
	var	R;

	R = S;
	S = 999;
	return R;
}

init_tests() {
	var	a = 12345, b = init();
	static var	SI00 = a / 5;

	TestID = 2700;

	test(GI00, 250);
	test(a, 12345);
	test(b, 5432);
	test(SI00, a/5);

	test(s_init(), 777);
	test(s_init(), 777);
}


///////////////////////////////////////////////////////////////////////////////
//
//	SIEVE TEST
//

sieve_test() {
	var	primes.100;
	var	i, j, top, ok;

	TestID = 2800;

	primes.0 = 2;
	top = 1;
	i = 3;
	while (top < 100) {
		ok = 1;
		for (j=0, top) {
			if (!(i%primes.j)) {
				ok = 0;
				break;
			}
		}
		if (ok) {
			primes.top = i;
			top += 1;
		}
		i += 2;
	}

	test(primes.(top-1), 541);
}


///////////////////////////////////////////////////////////////////////////////
//
//	THINGS THAT WENT WRONG WHEN THE COMPILER WAS CONSTRUCTED ...
//

ot1() return 1000;
ot2() return 2000;

o_test_1() {
	var	selector;

	TestID = 2900;

	selector = 0;
	test((selector-> ot1, ot2)(), 2000);
	selector = 1;
	test((selector-> ot1, ot2)(), 1000);
	selector = 0;
	test((!selector-> ot1, ot2)(), 1000);
	selector = 1;
	test((!selector-> ot1, ot2)(), 2000);

	selector = 0;
	test((selector-> ot1, ot2)(0, 1, 2, 3, 4, 5, 6, 7, 8, 9), 2000);
	selector = 1;
	test((selector-> ot1, ot2)(9, 8, 7, 6, 5, 4, 3, 2, 1, 0), 1000);
}


o_test_2() {
	var	a, slc;

	TestID = 3000;

	a = 100;

	slc = 1;
	a -= (slc-> 20, 30) + 10;
	test(a, 70);
	slc = 0;
	a += (slc-> 30, 20) + 10;
	test(a, 100);

	a = '9';
	a -= ('0' <= a && a <= '9'-> '0', 'A'-10);
	test(a, 9);
	a = 'F';
	a -= ('0' <= a && a <= '9'-> '0', 'A'-10);
	test(a, 15);
}


///////////////////////////////////////////////////////////////////////////////
//
//	HUGE STATIC TABLES CAUSED THE OPTIMIZER TO CRASH WHEN IT
//	WALKED DOWN THE N-LIST CREATED BY THE PARSER (STACK OVERFLOW).
//	THE FIX IS, OF COURSE, NOT TO FOLLOW NUMERIC LISTS...
//	[SEE OPT.C, OPT_CEXPR()]
//
o_test_3() {
	var	i;
	value	N = 0xa55a;

	static var	big_table = [
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,
	N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N,N
	];

	TestID = 3100;

	for (i=0, 512) if (big_table.i != N) break;
	test(big_table.(i-1), N);
}

///////////////////////////////////////////////////////////////////////////////
//
//	*INI* INSTRUCTIONS MUST NOT BE REVERSED IN ARGUMENT LISTS
//	OF FUNCTION CALLS. SEE *ORDER()* IN FILE EXPR.BSL.
//

ot4sbr(s1, s2, s3, t) {
	test(s1.0, t.0);
	test(s1.1, t.0);
	test(s1.2, t.0);
	test(s1.3, 0);

	test(s2.0, t.1);
	test(s2.1, t.1);
	test(s2.2, t.1);
	test(s2.3, 0);

	test(s3.0, t.2);
	test(s3.1, t.2);
	test(s3.2, t.2);
	test(s3.3, 0);
}


o_test_4() {
	TestID = 3200;

	ot4sbr("111", "222", "333", "123");
	ot4sbr("444", "555", "666", "456");
	ot4sbr("777", "888", "999", "789");
}


obscure_tests() {
	o_test_1();
	o_test_2();
	o_test_3();
	o_test_4();
}


///////////////////////////////////////////////////////////////////////////////
//
//	HERE THE FUN BEGINS
//

start() {
	oper_tests();
	prec_tests();
	assoc_test();
	cexpr_tests();
	sign_tests();
	addr_tests();
	func_tests();
	branch_tests();
	switch_tests();
	loop_tests();
	init_tests();
	sieve_test();
	obscure_tests();
	if (!ErrorC) {
		writen(TestN);
		writes(" TESTS PASSED\r\n");
	}
	halt 123-123;
	fail();
}
