///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: I/O-Streams
//	Copyright (C) 1995 Nils M. Holm
//
//	SOPEN ( NAME, MODE )
//
//	PURPOSE:	Open/create an I/O-Stream
//	RECEIVES:	NAME	- file to open
//			MODE	- open mode: see IOS.HDR
//	RETURNS:	A stream pointer, or ENDSTREAM in case
//			of an error
//
//	NOTES:		The file will be opened in compatibility mode.
//
//			The IOS-functions require the dynamic memory
//			manager, so MMINIT() must be called before
//			using any IOS-functions.
//

#include( blib.hdr)
#include( ../opt.hdr)
#include( ../sys/syscall.hdr)

#define( EXTERN(x),)
#include( ios.hdr)
#undefine(EXTERN)

extern	screate;

public sopen(name, mode) {
	var	ios, omode, fd;

	if (!(mode & S_RDWR) || mode & ~_S_MASK) return ENDSTREAM;

	switch (mode & S_RDWR) {
		case S_READ:	omode = 0; break;
		case S_WRITE:	omode = 1; break;
		case S_RDWR:	omode = 1; break;
	}

	if (mode & S_CREATE) {
		fd = sys_create(name);
	}
	else {
		fd = sys_open(name, omode);
	}
	if (fd < 0) return ENDSTREAM;

	if (mode & S_APPEND) sys_seek(fd, 0, 2);

	if (!(ios = screate(mode))) return ENDSTREAM;
	ios.ios_fd = fd;

	return ios;
}
