///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: I/O-Streams
//	Copyright (C) 1995 Nils M. Holm
//
//	SSYNC ( IOS )
//
//	PURPOSE:	Synchronize an I/O-Stream
//	RECEIVES:	IOS	- Pointer to the stream to sync
//	RETURNS:	0 = success, ENDSTREAM = lossage
//
//	NOTES:		The IOS-functions require the dynamic memory
//			manager, so MMINIT() must be called before
//			using any IOS-functions.
//

#include( blib.hdr)
#include( ../opt.hdr)
#include( ../sys/syscall.hdr)
#include( ../mem/mem.hdr)

#define( EXTERN(x),)
#include( ios.hdr)
#undefine(EXTERN)

public ssync(ios) {
	if (ios.ios_flags & S_ERROR) {
		ios.ios_ptr = ios.ios_end = ios.ios_buf;
		return ENDSTREAM;
	}
	if (ios.ios_flags & S_WRITE) {
		var	len = ios.ios_ptr - ios.ios_buf;

		if (len) {
			bpack(ios.ios_buf, ios.ios_buf, len);
			if (sys_write(ios.ios_fd, ios.ios_buf, len) != len) {
				ios.ios_flags |= S_ERROR;
				ios.ios_ptr = ios.ios_end = ios.ios_buf;
				return ENDSTREAM;
			}
		}
	}
	ios.ios_ptr = ios.ios_end = ios.ios_buf;
	ios.ios_ubuf = ENDSTREAM;
	return 0;
}
