///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: Dynamic Memory Management
//	Copyright (C) 1995 Nils M. Holm
//
//	MMINIT ( ARENA, SIZE )
//
//	PURPOSE:	Initialize the memory manager
//	RECEIVES:	ARENA	- pointer to arena
//			SIZE	- size of arena in words
//	RETURNS:	Nothing.
//
//	NOTES:		MMINIT() must be called before any
//			other memory function may be used.
//			The ARENA must not be de-allocated
//			while the memory manager is in use.
//
//	--------------------------------------------------
//
//	MALLOC ( SIZE )
//
//	PURPOSE:	Allocate a block
//	RECEIVES:	SIZE	- size of requested block
//	RETURNS:	A pointer to the allocated block,
//			or Null if no block with a sufficient
//			size exists.
//
//	--------------------------------------------------
//
//	MFREE ( PTR )
//
//	PURPOSE:	Release an allocated block, and
//			defragment the arena.
//	RECEIVES:	PTR	- pointer to the block to release
//	RETURNS:	0 = success, -1 = bad pointer;
//

#include( blib.hdr)
#include( ../opt.hdr)

value	USEDBIT = 0x8000;

var	Pool = 0;


public mminit(arena, size) {
	Pool = arena;
	Pool.0 = (size-1);
	Pool.(size-1) = 0;
}


public malloc(size) {
	var	mp = Pool, np;

	if (!Pool || !size || size & USEDBIT) return 0;

	size += 1;
	while (.mp) {
		if (!(.mp & USEDBIT) && .mp U>= size) {
			if (.mp == size) {
				.mp |= USEDBIT;
				return mp+1;
			}
			else {
				np = mp + size;
				.np = .mp - size;
				.mp = size | USEDBIT;
				return mp+1;
			}
		}
		else {
			mp += .mp & ~USEDBIT;
		}
	}

	return 0;
}


public mfree(ptr) {
	var	mp = Pool, head = 0;

	if (!Pool || !(.(ptr-=1) & USEDBIT)) return -1;

	.ptr &= ~USEDBIT;
	while (.mp) {
		if (!(.mp & USEDBIT)) {
			if (!head) head = mp;
		}
		else {
			if (head && mp U> ptr)
				break;
			else
				head = 0;
		}
		mp += .mp & ~USEDBIT;
	}
	.head = mp-head;

	return 0;
}
