///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: String Functions
//	Copyright (C) 1995 Nils M. Holm
//
//	BTOS ( VAL, BASE, LENGTH )
//
//	PURPOSE:	Convert binary to string.
//	RECEIVES:	VAL	- Value to convert.
//			BASE	- Numeric base for STRING.
//			LENGTH	- Minimum length of the resulting string.
//				When the string is shorter than LENGTH,
//				it will be padded with leading blanks.
//	RETURNS:	A pointer to a static buffer containing
//			the readable form of VAL.
//			BTOS() ignores the sign of VAL unless
//			BASE is negative. If BASE is -10, for
//			example, a leading minus will be
//			generated, if VAL is less than zero.
//			BASE must be in the range 2...16 or
//			-16...-2.
//
//	NOTES:	The length of the generated string must not exceed
//		32 characters !
//

#include( blib.hdr)
#include( ../opt.hdr)

public btos(val, base, length) {
	static	buf.33;
	var	p = 31, n, sign = 0;

	buf.32 = 0;
	if (base < 0) {
		base = -base;
		if (val<0) {
			val = -val;
			sign = -1;
		}
	}
	while (val) {
		n = val % base;
		buf.p = n + (n>9-> 'A'-10, '0');
		p -= 1;
		val /= base;
	}
	if (p == 31) {
		buf.p = '0'; p -= 1;
	}
	if (sign) {
		buf.p = '-'; p -= 1;
	}
	while (31-p < length) {
		buf.p = '\s'; p -= 1;
	}
	return buf+p+1;
}
