///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: String Functions
//	Copyright (C) 1995 Nils M. Holm
//
//	VFORMAT ( BUFFER, FORMAT, ARGV )
//
//	PURPOSE:	String formatting.
//	RECEIVES:	BUFFER	- where to place the formatted string.
//			FORMAT	- the description of the desired format.
//			ARGV	- a pointer to an argument vector holding
//				the objects to be formatted according to
//				the descriptions in FORMAT.
//	RETURNS:	A pointer to BUFFER.
//
//	DESCRIPTION:	VFORMAT() copies each character from FORMAT to BUFFER,
//			except for percent character (%). When a percent sign
//			is found, VFORMAT() takes the next argument from the
//			parameter list (starting with ARGS), and formats it
//			according to the description following '%'. The percent
//			sign itself is discarded. To insert a literal '%' into
//			buffer, use "%%".
//
//	%-DESCRIPTIONS:	Each %-description has the following format:
//
//				%[#][u]<type>
//
//			% (mandatory): introduces the description.
//			# (optional): a decimal number that specifies the
//				length of the field.
//			U (optional): when specified, it makes the following
//				numeric operand unsigned. U has no effect
//				when used with 's' or 'c'.
//			<type> (mandatory): a single character that specifies
//				the type of the current argument:
//				b - binary number
//				d - decimal number
//				o - octal number
//				x - hexadecimal number
//				c - character
//				s - string
//				w - white space
//
//			When a field length is given, numeric objects will
//			be right-justified, and strings will be right left-
//			justified. The length is ignored for characters.
//

#include( blib.hdr)
#include( ../opt.hdr)

extern	btos, strcpy, strlen;

public vformat(buffer, format, argv) {
	var	pbuffer = buffer;
	var	flen, sign;

	while (.format) {
		if (.format != '%') {
			.buffer = .format;
			buffer += 1;
			format += 1;
			loop;
		}
		format += 1;
		if (.format == '%') {
			.buffer = '%';
			buffer += 1;
			format += 1;
			loop;
		}
		sign = -1;
		flen = 0;
		while ('0' <= .format && .format <= '9') {
			flen = flen*10 + .format - '0';
			format += 1;
		}
		if (.format == 'u' || .format == 'U') {
			sign = 0;
			format += 1;
		}
		switch (.format) {
			case 0:
			format -= 1; break;

			case 'b': case 'B':
			strcpy(buffer, btos(.argv, sign-> -2, 2, flen));
			break;

			case 'd': case 'D':
			strcpy(buffer, btos(.argv, sign-> -10, 10, flen));
			break;

			case 'o': case 'O':
			strcpy(buffer, btos(.argv, sign-> -8, 8, flen));
			break;

			case 'x': case 'X':
			strcpy(buffer, btos(.argv, sign-> -16, 16, flen));
			break;

			case 'w': case 'W':
			case 's': case 'S': {
				var	l = 0;

				if (.format == 's' || .format == 'S') {
					l = strlen(.argv);
					strcpy(buffer, .argv);
					buffer += l;
				}
				while (l < flen) {
					.buffer = '\s';
					buffer += 1;
					l += 1;
				}
				.buffer = 0;
				break;
			}

			case 'c': case 'C':
			.buffer = .argv;
			buffer.1 = 0;
			break;

			default:
			.buffer = 0;
		}
		while (.buffer) buffer += 1;
		format += 1;
		argv += 1;
	}
	.buffer = 0;

	return pbuffer;
}
