///////////////////////////////////////////////////////////////////////////////
//
//	BSL RUNTIME LIBRARY TEST SUITE
//	Copyright (C) 1995 Nils M. Holm
//

#include( blib.hdr)

#include( ../mem/mem.hdr)
#include( ../str/str.hdr)
#include( ../ios/ios.hdr)
#include( ../sys/syscall.hdr)

// Where is your command interpreter ??
//
//var	COMMAND = "D:/OS2/MDOS/COMMAND.COM";
var	COMMAND = "C:/COMMAND.COM";


value	USEDBIT = 0x8000;


var	Err_count = 0;


fail(msg) {
	var	l, nl.3;

	l = pack(msg, msg);
	writepacked(SYSOUT, msg, l);
	nl.0 = '\r';
	nl.1 = '\n';
	nl.2 = 0;
	pack(nl, nl);
	writepacked(SYSOUT, nl, 2);
	Err_count += 1;
}


block_tests() {
	var	a, b, v.256;
	var	i;

	// BCMP() bounds
	//
	a = "abcdefghijklmnopqrstuvwxyz";
	if (bcmp(a, a, 27)) fail("bcmp-1");
	b = "abcdefghij................";
	if (bcmp(a, b, 10)) fail("bcmp-2");
	if (bcmp(a, b, 11) != 'k' - '.') fail("bcmp-3");

	// BCOPY() bounds
	//
	bcopy("ABCDEFGHIJKLMNOPQRSTUVWXYZ", v, 27);
	for (i=0, 26) {
		if (v.i != i+'A') fail("bcopy-1"); break;
	}
	if (v.26) fail("bcopy-2");
	bcopy("0123456789", v, 10);
	if (bcmp(v, "0123456789KLMNOPQRSTUVWXYZ", 27)) fail("bcopy-3");

	// BZERO() bounds
	//
	bset(v, 0, 10);
	if (bcmp(v, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0], 10)) fail("bset-1");
	bset(v, 0xFFF, 10);
	if (bcmp(v,
	[0xFFF, 0xFFF, 0xFFF, 0xFFF, 0xFFF, 0xFFF, 0xFFF, 0xFFF, 0xFFF, 0xFFF],
	10))
		fail("bset-2");
	if (bcmp(v+10, "KLMNOPQRSTUVWXYZ", 17)) fail("bset-3");
}


do_pack(a) {
	var	pa.16, ua.16, len = 0;

	while (a.len) len += 1;
	len += 1;
	bpack(a, pa, len);
	bunpack(pa, ua, len);
	if (bcmp(a, ua, len)) {
		fail("bpack/bunpack");
		return -1;
	}
	return 0;
}


pack_tests() {
	// BPACK() and BUNPACK() with all possible alignments
	// up to word size == 32 bits
	//
	if (do_pack("PACK THIS: 1")) return;
	if (do_pack("PACK THIS: 12")) return;
	if (do_pack("PACK THIS: 123")) return;
	if (do_pack("PACK THIS: 1234")) return;
}


scan_tests() {
	var	a;

	a = "..........x..........x";
	//   |         |          |
	//   a        a+10       a+21

	// SCAN() regions
	//
	if (bscan(a, '.', 1) != a) fail("bscan-1");
	if (bscan(a, 'x', 11) != a+10) fail("bscan-2");
	if (bscan(a, 'x', 10)) fail("bscan-3");
	if (bscan(a, 'x', 22) != a+10) fail("bscan-4");
	if (bscan(a, 'y', 23)) fail("bscan-4");

	// RSCAN() regions
	//
	if (brscan(a, '.', 5) != a+4) fail("brscan-1");
	if (brscan(a, 'x', 11) != a+10) fail("brscan-2");
	if (brscan(a, 'x', 10)) fail("brscan-3");
	if (brscan(a, 'x', 22) != a+21) fail("brscan-4");
	if (brscan(a, 'y', 23)) fail("brscan-4");
}


malloc_tests() {
	var	arena . 1024;
	var	p1, p2, p3;

	// Arena (free list) initialization
	//
	mminit(arena, 1024);
	if (.arena != 1023) fail("mminit-1");
	if (arena.1023) fail("mminit-2");

	// Allocation of greatest possible block
	//
	if (!(p1 = malloc(1022))) {
		fail("malloc-1"); return;
	}
	if (.arena != (1023 | USEDBIT)) fail("malloc-2");
	if (p1 != arena+1) fail("malloc-3");

	// De-allocation of a single large block
	//
	if (mfree(p1)) fail("mfree-1");
	if (.arena != 1023) fail("mfree-2");
	if (arena.1023) fail("mfree-3");

	// Allocation of serveral small blocks
	//
	if (!(p1 = malloc(256))) fail("malloc-4");
	if (arena.257 != 766) fail("malloc-5");
	if (!(p2 = malloc(256))) fail("malloc-6");
	if (!(p3 = malloc(256))) fail("malloc-7");

	// Heap de-fragmentation while releasing
	// small blocks
	//
	if (mfree(p1)) fail("mfree-4");
	if (p1.(-1) != 257) fail("mfree-5");
	if (mfree(p3)) fail("mfree-6");
	if (p3.(-1) != 509) fail("mfree-7");
	if (mfree(p2)) fail("mfree-8");

	// Everything released??
	//
	if (.arena != 1023) fail("mfree-9");
}


memory_tests() {
	block_tests();	// BCMP, BCOPY, BZERO
	pack_tests();	// MPACK, MUNPACK
	scan_tests();	// SCAN, RSCAN
	malloc_tests();	// MMINIT, MALLOC, MFREE
}


conv_tests() {
	var	s, i, j;

	// Conversion of decimal, hex, octal, and
	// binary numbers to strings
	//
	s = btos(1234, 10, 0);
	if (bcmp(s, "1234", 5)) fail("btos-1");
	s = btos(0xfff, 16, 0);
	if (bcmp(s, "FFF", 4)) fail("btos-2");
	s = btos(03333, 8, 0);
	if (bcmp(s, "3333", 5)) fail("btos-3");
	s = btos(0b101010101010, 2, 0);
	if (bcmp(s, "101010101010", 13)) fail("btos-4");

	// Unsigned conversion (positive base)
	//
	s = btos(-1, 16, 0);
	while (.s) {
		if (.s != 'F') {
			fail("btos-5"); break;
		}
		s += 1;
	}

	// Signed conversion (negative base)
	//
	s = btos(-1, -10, 0);
	if (bcmp(s, "-1", 3)) fail("btos-6");

	// Conversion of Zero
	//
	s = btos(0, 10, 0);
	if (.s != '0' || s.1 != 0) fail("btos-7");

	// Padding test (leading spaces)
	//
	for (i=1, 32) {
		s = btos(1, 10, i);
		for (j=0, i-1) {
			if (s.j != '\s') {
				fail("btos-8"); break;
			}
		}
		if (s.j != '1') {
			fail("btos-9"); break;
		}
	}

	// String evaluation tests
	//
	if (stob("    1234", 10) != 1234) fail("stob-1");
	if (stob("     EeE", 16) != 0xEEE) fail("stob-2");
	if (stob("   07654", 8) != 07654) fail("stob-3");
	if (stob("10100101", 2) != 0b10100101) fail("stob-4");

	// Sign recognition tests
	//
	if (stob("   -4321", 10) != -4321) fail("stob-5");
	if (stob("   +4321", 10) != +4321) fail("stob-6");

	// Leading TABs
	//
	if (stob("\t\t5555", 10) != 5555) fail("stob-7");

	// Non-numeric input
	//
	if (stob("X", 10)) fail("stob-8");
}


rx_tests() {
	var	pattern . 100, test . 100;
	var	i, t;

	// Literal matches
	//
	rxcomp("abcdef", pattern);
	if (!rxmatch(pattern, "abcdef")) fail("rxmatch-1");
	if (!rxmatch(pattern, ".....abcdef.....")) fail("rxmatch-2");
	if (rxmatch(pattern, "abcde")) fail("rxmatch-3");

	// Class matches
	//
	rxcomp("[a-f]", pattern);
	for (i='\s', '~'+1) {
		test.0 = i; test.1 = 0;
		if ('a' <= i && i <= 'f' && !rxmatch(pattern, test)) {
			fail("rxmatch-4"); break;
		}
		else if ((i < 'a' || i > 'f') && rxmatch(pattern, test)) {
			fail("rxmatch-5"); break;
		}
	}

	// Reverse class matches
	//
	rxcomp("[^a-f]", pattern);
	for (i='\s', '~'+1) {
		test.0 = i; test.1 = 0;
		if ('a' <= i && i <= 'f' && rxmatch(pattern, test)) {
			fail("rxmatch-6"); break;
		}
		else if ((i < 'a' || i > 'f') && !rxmatch(pattern, test)) {
			fail("rxmatch-7"); break;
		}
	}

	// Match the beginning of line
	//
	rxcomp("^abcdef", pattern);
	if (!rxmatch(pattern, "abcdef")) fail("rxmatch-8");
	if (rxmatch(pattern, "xabcdef")) fail("rxmatch-9");

	// Match the end of line
	//
	rxcomp("abcdef$", pattern);
	if (!rxmatch(pattern, "xxxabcdef")) fail("rxmatch-10");
	if (rxmatch(pattern, "abcdefx")) fail("rxmatch-11");

	// Match a placeholder
	//
	rxcomp("abc.ef", pattern);
	t = "abcdef";
	for (i='\s', '~'+1) {
		t.3 = i;
		if (!rxmatch(pattern, t)) {
			fail("rxmatch-12"); break;
		}
	}
}


rx2_tests() {
	var	pattern . 100;

	// Longest match (including zero)
	//
	rxcomp("aaa.*xxx", pattern);
	if (!rxmatch(pattern, "aaa0123456789xxx")) fail("rxmatch-13");
	if (!rxmatch(pattern, "aaa.xxx")) fail("rxmatch-14");
	if (!rxmatch(pattern, "aaaxxx")) fail("rxmatch-15");
	if (rxmatch(pattern, "aaaxx")) fail("rxmatch-16");
	rxcomp("ab*c", pattern);
	if (!rxmatch(pattern, "ac")) fail("rxmatch-17");
	if (!rxmatch(pattern, "abc")) fail("rxmatch-18");
	if (!rxmatch(pattern, "abbbc")) fail("rxmatch-19");

	// Longest match (excluding zero)
	//
	rxcomp("ab+c", pattern);
	if (!rxmatch(pattern, "abbbc")) fail("rxmatch-20");
	if (!rxmatch(pattern, "abc")) fail("rxmatch-21");
	if (rxmatch(pattern, "ac")) fail("rxmatch-22");

	// Optional match
	//
	rxcomp("^aaab-ccc$", pattern);
	if (!rxmatch(pattern, "aaaccc")) fail("rxmatch-23");
	if (!rxmatch(pattern, "aaabccc")) fail("rxmatch-24");
	if (rxmatch(pattern, "aaabbccc")) fail("rxmatch-25");

	// Empty RE
	//
	rxcomp("^$", pattern);
	if (!rxmatch(pattern, "")) fail("rxmatch-26");
	if (rxmatch(pattern, "x")) fail("rxmatch-27");

	// Escape test
	//
	rxcomp("^\\^\\[abc\\]\\*\\+\\-\\\\\\.\\$$", pattern);
	if (!rxmatch(pattern, "^[abc]*+-\\.$")) fail("rxmatch-28");

	// Escaped class test
	//
	rxcomp("[D\\-F]", pattern);
	if (!rxmatch(pattern, "-")) fail("rxmatch-29");
	if (rxmatch(pattern, "E")) fail("rxmatch-30");
}


str_tests() {
	var	a, i, v . 100;

	// ASCIZ string length
	//
	if (strlen("")) fail("strlen-1");
	if (strlen("abc") != 3) fail("strlen-2");
	if (strlen("abcdef\0abcdef") != 6) fail("strlen-3");

	// STRCMP() bounds
	//
	a = "0123456789ABCDEF";
	if (strcmp(a, "0123456789ABCDEF")) fail("strcmp-1");
	if (!strcmp(a, "0123456789ABCDEFG")) fail("strcmp-2");
	if (!strcmp(a, "0123456789ABCDE")) fail("strcmp-3");

	a = "x.........y..........y";
	//   |         |          |
	//   a        a+10       a+21

	// STRCHR() value
	//
	if (strchr(a, 'x') != a) fail("strchr-1");
	if (strchr(a, 'y') != a+10) fail("strchr-2");
	if (strchr(a, 'z')) fail("strchr-3");
	if (strchr(a, '\0')) fail("strchr-4");

	// STRRCHR() value
	//
	if (strrchr(a, 'x') != a) fail("strrchr-1");
	if (strrchr(a, 'y') != a+21) fail("strrchr-2");
	if (strrchr(a, 'z')) fail("strrchr-3");
	if (strrchr(a, '\0')) fail("strrchr-4");

	// STRCPY() bounds
	//
	v.7 = 12345;
	strcpy(v, "abcdef");
	if (strcmp(v, "abcdef")) fail("strcpy-1");
	if (v.7 != 12345) fail("strcpy-2");
	v.1 = 4321;
	strcpy(v, "");
	if (.v) fail("strcpy-3");
	if (v.1 != 4321) fail("strcpy-4");

	// STRCAT() bounds
	//
	strcpy(v, "abc");
	v.7 = 777;
	strcat(v, "def");
	if (strcmp(v, "abcdef")) fail("strcat-1");
	if (v.7 != 777) fail("strcat-2");
	strcat(v, "\0");
	if (strcmp(v, "abcdef")) fail("strcat-3");
	if (v.7 != 777) fail("strcat-4");

	// Case conversion
	//
	for (i='\s', '~'+1) {
		if ('a' <= i && i <= 'z' && toupper(i) != i) loop;
		if ('A' <= i && i <= 'Z' && tolower(i) != i) loop;
		if (toupper(i) == i) loop;
		if (tolower(i) == i) loop;
		fail("toupper/tolower");
		break;
	}
}


fmt_tests() {
	var	s . 100;

	// Basic format descriptor types
	//
	strfmt(s, "<%s>", "Hello, World!");
	if (strcmp(s, "<Hello, World!>")) fail("strfmt-1");
	strfmt(s, "%d", 12345);
	if (strcmp(s, "12345")) fail("strfmt-2");
	strfmt(s, "%b", 0b1010110101);
	if (strcmp(s, "1010110101")) fail("strfmt-3");
	strfmt(s, "%o", 03777);
	if (strcmp(s, "3777")) fail("strfmt-4");
	strfmt(s, "%x", 0xFFF);
	if (strcmp(s, "FFF")) fail("strfmt-5");
	strfmt(s, "XX%cXX", 'X');
	if (strcmp(s, "XXXXX")) fail("strfmt-6");
	strfmt(s, "%%");
	if (strcmp(s, "%")) fail("strfmt-7");

	// Signed/unsigned numeric formats
	//
	strfmt(s, "%x", -1);
	if (strcmp(s, "-1")) fail("strfmt-8");
	strfmt(s, "%ux", -1);
	if (.s != 'F') fail("strfmt-9");

	// Padding fields
	//
	strfmt(s, "'%10d'", -1235);
	if (strcmp(s, "'     -1235'")) fail("strfmt-10");
	strfmt(s, "%1d", 1234);
	if (strcmp(s, "1234")) fail("strfmt-11");
	strfmt(s, "<%20s>", "STRING-TEST");
	if (strcmp(s, "<STRING-TEST         >")) fail("strfmt-12");
	strfmt(s, "<%2s>", "STRING-TEST");
	if (strcmp(s, "<STRING-TEST>")) fail("strfmt-13");
}


fmt2_tests() {
	var	s . 100, s2;
	var	n, n2, i;

	// Basic format descriptors
	//
	strscan("Scan me", "%s", s);
	if (strcmp(s, "Scan me")) fail("strscan-1");
	strscan("12345", "%d", @n);
	if (n != 12345) fail("strscan-2");
	strscan("1011001101", "%b", @n);
	if (n != 0b1011001101) fail("strscan-3");
	strscan("5555", "%o", @n);
	if (n != 05555) fail("strscan-4");
	strscan("ABCD", "%x", @n);
	if (n != 0xABCD) fail("strscan-5");
	strscan("xyz", "%c", @n);
	if (n != 'x') fail("strscan-6");
	strscan("0123456789ABCDEF", "%[0-9]", s);
	if (strcmp(s, "0123456789")) fail("strscan-7");
	strscan("%XYZ", "%%%s", s);
	if (strcmp(s, "XYZ")) fail("strscan-8");

	// Leading fillers and sign tests
	//
	strscan("-123", "%d", @n);
	if (n != -123) fail("strscan-9");
	strscan("\s\s\sEEE", "\s%x", @n);
	if (n != 0xEEE) fail("strscan-10");
	strscan("\t\t\t333", "\t%o", @n);
	if (n != 0333) fail("strscan-11");
	strscan("ZZZZZZZ999", "Z%d", @n);
	if (n != 999) fail("strscan-12");

	// Length tests
	//
	strscan("11111111", "%4b", @n);
	if (n != 0b1111) fail("strscan-13");
	strscan("1234567890", "%7s%d", s, @n);
	if (n != 890) fail("strscan-14");

	// Class tests
	//
	s2 = s+50;
	strscan("ABCABCDDD535", "%[ABC]%4s%d", s, s2, @n);
	if (strcmp(s2, "DDD5") || n != 35) fail("strscan-15");
	i = strscan("-x-x-x-x-123-x-x-x-x-x456", "%[-x]%d%[x-]%d",
	s, @n, s2, @n2);
	if (n != 123 || n2 != 456) fail("strscan-16");
	if (i != 4) fail("strfmt-17");
}


string_tests() {
	conv_tests();	// BTOS, STOB
	rx_tests();	// RXCOMP, RXMATCH
	rx2_tests();	// RXCOMP, RXMATCH
	str_tests();	// STRLEN, STRCMP, STRCHR, STRRCHR,
			// STRCPY, STRCAT
	fmt_tests();	// STRFMT (VFORMAT)
	fmt2_tests();	// STRSCAN (VSCAN)
}


file_test() {
	var	fd;
	var	testfile = "test.000";

	// Remove left over test file
	//
	sys_remove(testfile);

	// Attempt to open a non-existant file
	//
	if ((fd = sys_open(testfile, 0)) >= 0) {
		fail("sys_open-1"); sys_close(fd);
	}

	// Create a new file, then close it again
	//
	if ((fd = sys_create(testfile)) < 0) {
		fail("sys_create-1");
	}
	if (sys_close(fd) < 0) fail("sys_close-1");

	// Open an existing file in different modes
	//
	if ((fd = sys_open(testfile, 0)) < 0) {
		fail("sys_open-1");
	}
	if (sys_close(fd) < 0) fail("sys_close-2");

	if ((fd = sys_open(testfile, 1)) < 0) {
		fail("sys_open-2");
	}
	if (sys_close(fd) < 0) fail("sys_close-3");

	if ((fd = sys_open(testfile, 2)) < 0) {
		fail("sys_open-3");
	}
	if (sys_close(fd) < 0) fail("sys_close-4");

	// Open a file with an invalid mode
	//
	if ((fd = sys_open(testfile, 99)) >= 0) {
		fail("sys_open-4");
		sys_close(fd);
	}

	// Remove the test file
	//
	if (sys_remove(testfile) < 0) fail("sys_remove-1");

	// Attempt to remove a non-existant file
	//
	if (sys_remove(testfile) >= 0) fail("sys_remove-2");
}


file2_tests(testfile, gap) {
	value	TFSIZE = 512;
	var	fd;
	var	pattern.128, buf.128;
	var	i, j, n;

	// Kill left over file
	//
	sys_remove(testfile);

	// Create test file
	if ((fd = sys_create(testfile)) < 0) {
		fail("sys_create-2"); return 0;	// GIVE UP !!!
	}

	// Write a test pattern into a file
	// AAA...AAA to ZZZ...ZZZ
	// Insert dots a GAP.
	//
	n = 'A';
	for (i=0, TFSIZE) {
		if (i == gap) {
			for (j=0, 128) pattern.j = '.';
		}
		else {
			for (j=0, 128) pattern.j = n;
		}
		if ((n += 1) > 'Z') n = 'A';
		bpack(pattern, pattern, 128);
		if (sys_write(fd, pattern, 128) != 128) {
			fail("sys_write-1"); break;
		}
	}

	// Seek to the beginning of the test image
	//
	if (sys_seek(fd, 0, 0) < 0) fail("sys_seek-1");

	// Read back the test image checking for correct data.
	//
	n = 'A';
	for (i=0, TFSIZE) {
		if (sys_read(fd, buf, 128) != 128) {
			fail("sys_read-1"); break;
		}
		bunpack(buf, pattern, 128);
		if (i == gap) {
			for (j=0, 128) {
				if (pattern.j != '.') {
					fail("sys_read-2 (1)"); break;
				}
			}
		}
		else {
			for (j=0, 128) {
				if (pattern.j != n) {
					fail("sys_read-2 (2)"); break;
				}
			}
		}
		if (j != 128) break;
		if ((n += 1) > 'Z') n = 'A';
	}

	return fd;
}


file3_tests(fd, testfile, gap) {
	var	buf.128, pattern.128;
	var	i;

	// Seek to the gap.
	//
	if (sys_seek(fd, gap*128, 0) < 0) fail("sys_seek-2 (1)");

	// Check if the seek actually moved to the gap.
	//
	if (sys_read(fd, buf, 128) != 128) fail("sys_read-3");
	bunpack(buf, pattern, 128);
	for (i=0, 128) {
		if (pattern.i != '.') {
			fail("sys_seek-2 (2)"); break;
		}
	}

	// Seek to the gap again and overwrite it with stars
	//
	if (sys_seek(fd, gap*128, 0) < 0) fail("sys_seek-3");
	for (i=0, 128) pattern.i = '*';
	bpack(pattern, pattern, 128);
	if (sys_write(fd, pattern, 128) != 128) fail("sys_write-2");

	// Check if the gap has been overwritten successfully.
	//
	if (sys_seek(fd, gap*128, 0) < 0) fail("sys_seek-3 (1)");
	if (sys_read(fd, buf, 128) != 128) fail("sys_read-4");
	bunpack(buf, pattern, 128);
	for (i=0, 128) {
		if (pattern.i != '*') {
			fail("sys_seek-3 (2)"); break;
		}
	}

	// Check seek to EOF
	//
	if (sys_seek(fd, 0, 2) < 0) fail("sys_seek-4");
	if (sys_read(fd, buf, 1) != 0) fail("sys_seek-5");

	if (sys_close(fd) < 0) fail("sys_close-5");
	if (sys_remove(testfile) < 0) fail("sys_remove-3");
}


file4_tests() {
	var	test1 = "test.004",
		test2 = "test.04b";
	var	fd;

	if ((fd = sys_create(test1)) < 0) {
		fail("sys_create-2"); return;	// GIVE UP !!!
	}
	if (sys_close(fd) < 0) fail("sys_close-6");
	if ((fd = sys_create(test2)) < 0) {
		sys_remove(test1);
		fail("sys_create-3"); return;	// GIVE UP !!!
	}
	if (sys_close(fd) < 0) fail("sys_close-6");

	if (sys_rename(test1, test2) >= 0) fail("sys_rename-1");
	if (sys_remove(test2) < 0) fail("sys_remove-4");
	if (sys_rename(test1, test2) < 0) fail("sys_rename-2");
	if (sys_rename(test2, test1) < 0) fail("sys_rename-3");

	if (sys_remove(test2) >= 0) fail("sys_rename-4");

	sys_remove(test1);
}


dir_tests() {
	var	testdir = "TESTDIR",
		testfile = "TESTDIR/test.000";
	var	path.128, i, fd;

	// Remove left over stuff
	//
	sys_remove(testfile);
	sys_rmdir(testdir);

	// Directory creation
	//
	if (sys_mkdir(testdir) < 0) fail("sys_mkdir-1");
	if (sys_mkdir(testdir) >= 0) fail("sys_mkdir-2");
	if ((fd = sys_create(testfile)) < 0)
		fail("sys_mkdir-2");
	else
		sys_close(fd);

	// Removal of directory containing files
	//
	if (sys_rmdir(testdir) >= 0) fail("sys_rmdir-1");

	// Removement of an empty directory
	//
	if (sys_remove(testfile) < 0) fail("sys_remove-5");
	if (sys_rmdir(testdir) < 0) fail("sys_rmdir-2");

	// Determine current working directory
	if (sys_getcwd(path) < 0) fail("sys_getcwd-1");
	i = 0;
	while (path.i && i<128) i += 1;
	if (i >= 128) fail("sys_getcwd-2");
}


spawn_test() {
	var	i, arg.2;

	// Test module exit codes
	//
	arg.1 = 0;
	for (i='0', '9'+1) {
		.arg = i;
		if (sys_spawn("./lt.exe", arg) != i) {
			fail("sys_spawn-1"); break;
		}
	}

	// Attempt to spawn non-existant module
	//
	if (sys_spawn("...", 0) >= 0) fail("sys_spawn-2");
}


system_tests() {
	value	GAP = 333;
	var	testfile = "test.002", fd;

	file_tests();		// SYS_CREATE, SYS_OPEN, SYS_CLOSE,
				// SYS_REMOVE
	fd = file2_tests(testfile, GAP);
				// SYS_CREATE, SYS_OPEN, SYS_READ,
				// SYS_WRITE, SYS_SEEK
	file3_tests(fd, testfile, GAP);
				// SYS_SEEK, SYS_READ, SYS_WRITE,
				// SYS_CLOSE, SYS_REMOVE
	file4_tests();		// SYS_RENAME
	spawn_test();		// SYS_SPAWN
	dir_tests();		// SYS_MKDIR, SYS_CHDIR, SYS_RMDIR,
				// SYS_GETCWD
}


ios_tests() {
	var	ios;
	var	testfile = "test.ios";

	sys_remove(testfile);

	// IOStream allocation
	//
	if ((ios = screate(S_READ)) == 0) fail("screate-1");
	if (.(ios-1) != (IOSTREAM+1 | USEDBIT)) fail("screate-2");
	if (.(ios.ios_buf-1) != (BUFLEN+1 | USEDBIT)) fail("screate-3");
	mfree(ios.ios_buf);
	mfree(ios);

	// IOStream open with invalid mode
	//
	if ((ios = sopen(testfile, -1)) != ENDSTREAM) fail("sopen-1");
	if ((ios = sopen(testfile, 0)) != ENDSTREAM) fail("sopen-2");

	// Attempt to open non-existant file
	//
	if ((ios = sopen(testfile, S_READ)) != ENDSTREAM) fail("sopen-3");
	if ((ios = sopen(testfile, S_WRITE)) != ENDSTREAM) fail("sopen-4");

	// File creation
	//
	if ((ios = sopen(testfile, S_WRITE|S_CREATE)) == ENDSTREAM)
		fail("sopen-5");
	if (sclose(ios)) fail("sclose-1");
	if ((ios = sopen(testfile, S_READ|S_CREATE)) == ENDSTREAM)
		fail("sopen-6");
	if (sclose(ios)) fail("sclose-2");

	// Opening an existing file
	//
	if ((ios = sopen(testfile, S_READ)) == ENDSTREAM) fail("sopen-7");
	if (sclose(ios)) fail("sclose-3");
	if ((ios = sopen(testfile, S_WRITE)) == ENDSTREAM) fail("sopen-8");
	if (sclose(ios)) fail("sclose-4");

	sys_remove(testfile);
}


ios2_tests() {
	var	ios, testfile = "test.io2";
	var	i;

	if ((ios = sopen(testfile, S_WRITE|S_CREATE)) == ENDSTREAM) {
		fail("sopen-9"); return;	// GIVE UP !!!
	}

	// IOStream correctly initialized?
	//
	if (	ios.ios_fd < 3				||
		ios.ios_flags != (S_WRITE|S_CREATE)	||
		ios.ios_bsize != BUFLEN			||
		ios.ios_ubuf  != ENDSTREAM		||
		ios.ios_buf   != ios.ios_ptr		||
		ios.ios_ptr   != ios.ios_end		)
			fail("sopen-10");

	// Pump some data into the stream, then check
	// for correct buffering
	//
	for (i='A', 'Z'+1) #sputc(ios, i);
	if (ios.ios_ptr - ios.ios_buf != 'Z'-'A'+1) fail("#sputc-1");
	for (i='A', 'Z'+1) {
		if (ios.ios_buf.(i-'A') != i) {
			fail("#sputc-2"); break;
		}
	}

	// Check IOS synchronisation
	//
	ssync(ios);
	if (ios.ios_buf != ios.ios_ptr) fail("ssync-1");

	if (sclose(ios)) fail("sclose-5");

	// Check stream-->file link
	//
	if ((ios = sopen(testfile, S_READ)) == ENDSTREAM) {
		fail("sopen-11"); return;	// GIVE UP !!!
	}
	for (i='A', 'Z'+1) {
		if (#sgetc(ios) != i) {
			fail("#sgetc-1"); break;
		}
	}
	if (#sgetc(ios) != ENDSTREAM) fail("#sgetc-2");

	// Check error recognition
	//
	if (!(ios.ios_flags & S_ERROR)) fail("smore-1");

	if (sclose(ios)) fail("sclose-6");

	sys_remove(testfile);
}


ios3_tests() {
	value	NREC = 512;
	var	ios, testfile = "test.io3";
	var	line.129;
	var	i, j, n, pass;

	if ((ios = sopen(testfile, S_WRITE|S_CREATE)) == ENDSTREAM) {
		fail("sopen-12"); return;	// GIVE UP !!!
	}

	// Fill test image with lines containing aaa...aaa to zzz...zzz
	//
	n = 'a';
	for (i=0, NREC) {
		for (j=0, 127) line.j = n;
		line.j = '\n';
		line.(j+1) = 0;
		if ((n += 1) > 'z') n = 'a';
		sputl(ios, line);
	}

	// Close test image, then reopen it for reading
	//
	if (sclose(ios)) fail("sclose-7");

	if ((ios = sopen(testfile, S_READ)) == ENDSTREAM) {
		fail("sopen-13"); return;	// GIVE UP !!!
	}

	// Read back test pattern, rewind, read back again
	//
	for (pass = 0, 2) {
		n = 'a';
		for (i=0, NREC) {
			sgetl(ios, line, 129);
			for (j=0, 127) {
				if (line.j != n) {
					fail(pass-> "sseek-2", "sgetl-1");
					break;
				}
			}
			if (j < 127) break;
			if (line.j != '\n' || line.(j+1)) {
				fail(pass-> "sseek-3", "sgetl-2");
				break;
			}
			if ((n += 1) > 'z') n = 'a';
		}
		if (!pass) {
			if (sseek(ios, 0, SEEK_SET) == ENDSTREAM)
				fail("sseek-1");
		}
	}

	// Random seek
	//
	if (sseek(ios, 4672, SEEK_SET) == ENDSTREAM) fail("sseek-4");
	sgetl(ios, line, 129);
	for (i=0, 63) {
		if (line.i != 'k') {
			fail("sseek-5 (1)"); break;
		}
	}
	if (line.i != '\n' || line.(i+1)) fail("sseek-5 (2)");

	// Seek to EOF
	//
	sseek(ios, 0, SEEK_END);
	if (#seof(ios)) fail("sseek-6");
	#sgetc(ios);
	if (!#seof(ios)) fail("sseek-7");
	sclose(ios);

	// SGETL() buffer overflow before LF
	//
	if ((ios = sopen(testfile, S_WRITE|S_CREATE)) ==  ENDSTREAM) {
		fail("sopen-14"); return;	// GIVE UP !!!
	}
	for (i=0, 65) #sputc(ios, 'X');
	#sputc(ios, '\n');
	if (#seof(ios)) fail("#sputc-3");
	sclose(ios);
	if ((ios = sopen(testfile, S_READ)) ==  ENDSTREAM) {
		fail("sopen-15"); return;	// GIVE UP !!!
	}
	sgetl(ios, line, 65);
	for (i=0, 64) {
		if (line.i != 'X') {
			fail("sgetl-3"); break;
		}
	}
	if (line.i) fail("sgetl-4");

	sclose(ios);
	sys_remove(testfile);
}


var	RULER =
	"0----|----1----|----2----|----3----|----4----|----5----|----6----|\n"
	;

//	WRITE TEST PATTERNS TO STDOUT
//
std_out(pattern) {
	var	arena . 1024;

	mminit(arena, 1024);
	stdopen();

	if (pattern == '0') {
		var	line.66;
		var	i, j;

		for (i=0, 64) {
			for (j=0, 64) line.j = i+'!';
			line.j = '\n';
			line.(j+1) = 0;
			sputl(STDOUT, line);
		}
	}

	else if (pattern == '1') {
		var	i;

		swritef(STDOUT, RULER);
		swritef(STDOUT,
	"|AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA|\n"
		);
		swritef(STDOUT, "|%64s|\n",
		"BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB");
		swritef(STDOUT, "|%32s%s|\n", "",
		"CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC");
		swritef(STDOUT, "|%32b%32ub|\n", 0xABCD, 0xABCD);
		swritef(STDOUT, "|%16o%16uo%16x%16ux|\n", 0xFEDC, 0xFEDC,
		0xFEDC, 0xFEDC);
		swritef(STDOUT, "|%16d%16ud%16d%16ud|\n", 9999, 9999,
		-9999, -9999);
		#sputc(STDOUT, '|');
		for (i=0, 8) {
			swritef(STDOUT, "%c%c%c%c%c%c%c%c",
			'A', 'B', 'C', 'D', '0', '1', '\s', '*');
		}
		sputl(STDOUT, "|\012|");
		for (i=0, 8) {
			swritef(STDOUT, "%%%%%%%%%%%%%%%%");
		}
		sputl(STDOUT, "|\012");
		swritef(STDOUT, RULER);
	}

	stdclose();
}


//	READ TEST PATTERNS FROM STDIN, AND COMPARE THEM
//	WITH THE ONES WRITTEN BY STD_OUT()
//
std_in(pattern) {
	var	arena . 1024;
	var	line.68;
	var	fail0 = "STDIO-0",
		fail1 = "STDIO-1";
	var	i;

	mminit(arena, 1024);
	stdopen();

	if (pattern == '0') {
		var	j;

		for (i=0, 64) {
			sgetl(STDIN, line, 66);
			for (j=0, 64) {
				if (line.j != i+'!') {
					fail(fail0); return 1;
				}
			}
			if (line.j != '\n' || line.(j+1)) {
				fail(fail0); return 1;
			}
		}

		if (#seof(STDIN)) { fail(fail0); return 1; }
		#sgetc(STDIN);
		if (!#seof(STDIN)) { fail(fail0); return 1; }
	}

	else if (pattern == '1') {
		var	a1, a2, a3, a4;
		var	s1.16, s2.64, s3.16;

		sgetl(STDIN, line, 68);
		if (strcmp(line, RULER)) {
			fail(fail1); return 1;
		}

		sgetl(STDIN, line, 68);
		for (i=1, 65) if (line.i != 'A') {
			fail(fail1); return 1;
		}

		sgetl(STDIN, line, 68);
		for (i=1, 33) if (line.i != 'B') {
			fail(fail1); return 1;
		}
		for (i=i, 65) if (line.i != '\s') {
			fail(fail1); return 1;
		}

		sgetl(STDIN, line, 68);
		for (i=1, 33) if (line.i != '\s') {
			fail(fail1); return 1;
		}
		for (i=i, 65) if (line.i != 'C') {
			fail(fail1); return 1;
		}

		sreadf(STDIN, "|\s%b %b|\n", @a1, @a2);
		if (a1 != a2 || a1 != 0xABCD) {
			fail(fail1); return 1;
		}

		sreadf(STDIN, "| %o %o %x %x|\n", @a1, @a2, @a3, @a4);
		if (a1 != a2 || a2 != a3 || a3 != a4 || a1 != 0xFEDC) {
			fail(fail1); return 1;
		}

		sreadf(STDIN, "|\s%d\s%d\s%d\s%d|\n", @a1, @a2, @a3, @a4);
		if (a1 != a2 || a1 != 9999 || a3 != a4 || a3 != -9999) {
			fail(fail1); return 1;
		}

		sreadf(STDIN, "|%8s%48s%[A-D0-1 *]%s", s1, s2, s3);
		if (strcmp(s1, "ABCD01 *") || strcmp(s1, s3)) {
			fail(fail1); return 1;
		}

		sreadf(STDIN, "|%%%s", s1);
		if (strcmp(s1, "|\n")) {
			fail(fail1); return 1;
		}

		sgetl(STDIN, line, 68);
		if (strcmp(line, RULER)) {
			fail(fail1); return 1;
		}

		if (#seof(STDIN)) { fail(fail1); return 1; }
		#sgetc(STDIN);
		if (!#seof(STDIN)) { fail(fail1); return 1; }
	}

	stdclose();
	return 0;
}


stdio_tests() {
	var	r;

	//
	// Write test patterns, read them back in, and compare them.
	//
	// The test suite spawns itself to redirect its standard
	// I/O to a test file. If a test fails, all test files will
	// be kept.
	//

	if (sys_spawn(COMMAND, " /c .\\lt.exe O 0 > test.st1") < 0)
		fail("sys_spawn-3");
	if ((r = sys_spawn(COMMAND, " /c .\\lt.exe I 0 < test.st1")) < 0)
		fail("sys_spawn-4");

	if (sys_spawn(COMMAND, " /c .\\lt.exe O 1 > test.st2") < 0)
		fail("sys_spawn-5");
	if ((r |= sys_spawn(COMMAND, " /c .\\lt.exe I 1 < test.st2")) < 0)
		fail("sys_spawn-6");

	if (!r) {
		sys_remove("test.st1");
		sys_remove("test.st2");
	}

	Err_count += r;
}


stream_tests() {
	var	arena . 2048;

	mminit(arena, 2048);	// Required for Stream-I/O

	ios_tests();		// SCREATE, SOPEN, SCLOSE
	ios2_tests();		// SOPEN, SSYNC, SMORE, SCLOSE,
				// "#SGETC", "#SPUTC"
	ios3_tests();		// SOPEN, SCLOSE, SGETL, SPUTL,
				// SSEEK
	stdio_tests();		// STDOPEN, STDCLOSE, SSCAN
				// SFORMAT
}


start(argcount, argvec) {
	var	passed = "PASSED\r\n";

	if ('0' <= ..argvec && ..argvec <= '9') return ..argvec;

	if (..argvec == 'O') {
		std_out(.(.argvec+1));
		halt 0;
	}
	else if (..argvec == 'I') {
		halt std_in(.(.argvec+1));
	}

	memory_tests();
	string_tests();
	system_tests();
	stream_tests();
	if (!Err_count) {
		pack(passed, passed);
		writepacked(SYSOUT, passed, 8);
	}
	return 0;
}
